@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2018 Artifex Software, Inc.
+/* Copyright (C) 2001-2023 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 /* Ordered Dither Screen Creation Tool. */
 
@@ -23,6 +23,7 @@
 #include "string_.h"
 #include "gsmemory.h"
 #include "math_.h"
+#include "gp.h"
 
 #   define ALLOC(mem, size) (gs_alloc_bytes((gs_memory_t *)mem, size, "gen_ordered"))
 #   define FREE(mem, ptr)   gs_free_object((gs_memory_t *)mem, ptr, "gen_ordered")
@@ -34,6 +35,15 @@
 #   define EPRINTF(mem, str) errprintf((gs_memory_t *)mem, str)
 #   define EPRINTF1(mem, str, v1) errprintf((gs_memory_t *)mem, str, v1)
 #   define EPRINTF3(mem, str, v1, v2, v3) errprintf((gs_memory_t *)mem, str, v1, v2, v3)
+#   define _FILE gp_file
+#   define FOPEN(mem, v1, v2) gp_fopen(mem, v1, v2)
+#   define FCLOSE(fid) gp_fclose(fid)
+#   define FPRINTF(fid, str) gp_fprintf(fid, str)
+#   define FPRINTF1(fid, str, v1) gp_fprintf(fid, str, v1)
+#   define FPRINTF2(fid, str, v1, v2) gp_fprintf(fid, str, v1, v2)
+#   define FPRINTF6(fid, str, v1, v2, v3, v4, v5, v6) gp_fprintf(fid, str, v1, v2, v3, v4, v5, v6)
+#   define FWRITE(v1, v2, v3, fid) gp_fwrite(v1, v2, v3, fid)
+
 
 #endif /* defined GS_LIB_BUILD */
 
@@ -64,6 +74,14 @@ typedef unsigned char byte;
 #   define EPRINTF(mem, str) fprintf(stderr, str)
 #   define EPRINTF1(mem, str, v1) fprintf(stderr, str, v1)
 #   define EPRINTF3(mem, str, v1, v2, v3) fprintf(stderr, str, v1, v2, v3)
+#   define _FILE FILE
+#   define FOPEN(mem, v1, v2) fopen(v1, v2)
+#   define FCLOSE(fid) fclose(fid)
+#   define FPRINTF(fid, str) fprintf(fid, str)
+#   define FPRINTF1(fid, str, v1) fprintf(fid, str, v1)
+#   define FPRINTF2(fid, str, v1, v2) fprintf(fid, str, v1, v2)
+#   define FPRINTF6(fid, str, v1, v2, v3, v4, v5, v6) fprintf(fid, str, v1, v2, v3, v4, v5, v6)
+#   define FWRITE(v1, v2, v3, fid) fwrite(v1, v2, v3, fid)
 
 #endif /* ndef LIB_BUILD */
 
@@ -105,8 +123,8 @@ typedef struct htsc_dither_pos_s {
     int *locations;
 } htsc_dither_pos_t;
 
-static void htsc_determine_cell_shape(double *x, double *y, double *v, double *u,
-                               double *N, htsc_param_t params, void *mem);
+static void htsc_determine_cell_shape(int *x, int *y, int *v, int *u,
+                               int *N, htsc_param_t params, void *mem);
 static double htsc_spot_value(spottype_t spot_type, double x, double y);
 static int htsc_getpoint(htsc_dig_grid_t *dig_grid, int x, int y);
 static void htsc_setpoint(htsc_dig_grid_t *dig_grid, int x, int y, int value);
@@ -124,15 +142,16 @@ static int htsc_allocate_supercell(htsc_dig_grid_t *super_cell, int x, int y, in
                            htsc_dig_grid_t dot_grid, int N, int *S, int *H, int *L);
 static void htsc_tile_supercell(htsc_dig_grid_t *super_cell, htsc_dig_grid_t *dot_grid,
                  int x, int y, int u, int v, int N);
-void create_2d_gauss_filter(float *filter, int x_size, int y_size,
-                            float stdvalx, float stdvaly);
+int create_2d_gauss_filter(double *filter, int x_size, int y_size,
+                            double stdvalx, double stdvaly, gs_memory_t *mem);
 static int htsc_create_holladay_mask(htsc_dig_grid_t super_cell, int H, int L,
                                double gamma, htsc_dig_grid_t *final_mask);
 static int htsc_create_dither_mask(htsc_dig_grid_t super_cell,
                              htsc_dig_grid_t *final_mask, int verbose,
                              int num_levels, int y, int x, double vert_dpi,
                              double horiz_dpi, int N, double gamma,
-                             htsc_dig_grid_t dot_grid, htsc_point_t one_index);
+                             htsc_dig_grid_t dot_grid, htsc_point_t one_index,
+                             gs_memory_t *mem);
 static int htsc_create_nondithered_mask(htsc_dig_grid_t super_cell, int H, int L,
                           double gamma, htsc_dig_grid_t *final_mask);
 static int htsc_gcd(int a, int b);
@@ -142,11 +161,11 @@ static void htsc_matrix_vector_mult(htsc_matrix_t matrix_in, htsc_vector_t vecto
                    htsc_vector_t *vector_out);
 static int htsc_mask_to_tos(htsc_dig_grid_t *final_mask);
 #if RAW_SCREEN_DUMP
-static void htsc_dump_screen(htsc_dig_grid_t *dig_grid, char filename[]);
-static void htsc_dump_float_image(float *image, int height, int width, float max_val,
-                      char filename[]);
-static void htsc_dump_byte_image(byte *image, int height, int width, float max_val,
-                      char filename[]);
+static int htsc_dump_screen(htsc_dig_grid_t *dig_grid, char filename[], gs_memory_t *mem);
+static int htsc_dump_float_image(double *image, int height, int width, double max_val,
+                      char filename[], gs_memory_t *mem);
+static int htsc_dump_byte_image(byte *image, int height, int width, double max_val,
+                      char filename[], gs_memory_t *mem);
 #endif
 
 /* Initialize default values */
@@ -169,10 +188,10 @@ void htsc_set_default_params(htsc_param_t *params)
 }
 
 int
-htsc_gen_ordered(htsc_param_t params, int *S, htsc_dig_grid_t *final_mask)
+htsc_gen_ordered(htsc_param_t params, int *S, htsc_dig_grid_t *final_mask, gs_memory_t *mem)
 {
-    double num_levels;
-    double x=0.0, y=0.0, v=0.0, u=0.0, N=0.0;
+    int num_levels;
+    int x=0, y=0, v=0, u=0, N=0;
     htsc_vertices_t vertices;
     htsc_vector_t bin_center;
     htsc_point_t one_index = { 0., 0. };
@@ -195,7 +214,7 @@ htsc_gen_ordered(htsc_param_t params, int *S, htsc_dig_grid_t *final_mask)
 
     /* Figure out how many levels to dither across. */
     if (params.targ_quant_spec) {
-        num_levels = ROUND((double) params.targ_quant / N);
+        num_levels = ROUND((double) params.targ_quant / (double)N);
     } else {
         num_levels = 1;
     }
@@ -231,10 +250,12 @@ htsc_gen_ordered(htsc_param_t params, int *S, htsc_dig_grid_t *final_mask)
        on sequence or dot profile */
     code = htsc_create_dot_mask(&dot_grid, x, y, u, v, params.scr_ang, vertices);
     if (code < 0) {
-        return -1;
+        return code;
     }
 #if RAW_SCREEN_DUMP
-    htsc_dump_screen(&dot_grid, "mask");
+    code = htsc_dump_screen(&dot_grid, "mask", mem);
+    if (code < 0)
+        return code;
 #endif
     /* A sanity check */
     if (htsc_sumsum(dot_grid) != -N) {
@@ -255,7 +276,9 @@ htsc_gen_ordered(htsc_param_t params, int *S, htsc_dig_grid_t *final_mask)
                             params.spot_type, trans_matrix_inv);
 
 #if RAW_SCREEN_DUMP
-    htsc_dump_screen(&dot_grid, "dot_profile");
+    code = htsc_dump_screen(&dot_grid, "dot_profile", mem);
+    if (code < 0)
+        return code;
 #endif
     /* Allocate super cell */
     code = htsc_allocate_supercell(&super_cell, x, y, u, v, params.targ_size,
@@ -267,19 +290,21 @@ htsc_gen_ordered(htsc_param_t params, int *S, htsc_dig_grid_t *final_mask)
 
     /* Make a warning about large requested quantization levels with no -s set */
     if (params.targ_size == 1 && num_levels > 1) {
-        min_size = (int)ceil((double)params.targ_quant / N);
+        min_size = (int)ceil((double)params.targ_quant / (double)N);
         EPRINTF1(final_mask->memory, "To achieve %d quantization levels with the desired lpi,\n", params.targ_quant);
         EPRINTF1(final_mask->memory, "it is necessary to specify a SuperCellSize (-s) of at least %d.\n", min_size);
         EPRINTF(final_mask->memory, "Note that an even larger size may be needed to reduce pattern artifacts.\n");
         EPRINTF(final_mask->memory, "Because no SuperCellSize was specified, the minimum possible size\n");
         EPRINTF(final_mask->memory, "that can approximate the requested angle and lpi will be created\n");
-        EPRINTF1(final_mask->memory, "with %d quantization levels.\n", (int) N);
+        EPRINTF1(final_mask->memory, "with %d quantization levels.\n", N);
     }
 
     /* Go ahead and fill up the super cell grid with our growth dot values */
     htsc_tile_supercell(&super_cell, &dot_grid, x, y, u, v, N);
 #if RAW_SCREEN_DUMP
-    htsc_dump_screen(&super_cell, "super_cell_tiled");
+    code = htsc_dump_screen(&super_cell, "super_cell_tiled", mem);
+    if (code < 0)
+        return code;
 #endif
     /* If we are using the Holladay grid (non dithered) then we are done. */
     if (params.holladay) {
@@ -292,17 +317,17 @@ htsc_gen_ordered(htsc_param_t params, int *S, htsc_dig_grid_t *final_mask)
             /* Dont allow nonsense settings */
             if (num_levels * N > super_cell.height * super_cell.width) {
                 EPRINTF3(final_mask->memory,
-                         "Notice, %3.0lf quantization levels not possible with super cell of %d by %d\n",
+                         "Notice, %d quantization levels not possible with super cell of %d by %d\n",
                          num_levels, super_cell.height, super_cell.width);
-                num_levels = ROUND((super_cell.height * super_cell.width) / N);
-                EPRINTF1(final_mask->memory, "Reducing dithering quantization to %3.0lf\n",
+                num_levels = ROUND((super_cell.height * super_cell.width) / (double)N);
+                EPRINTF1(final_mask->memory, "Reducing dithering quantization to %d\n",
                          num_levels);
                 EPRINTF1(final_mask->memory, "For an effective quantization of %d\n",
                          super_cell.height * super_cell.width);
             }
             code = htsc_create_dither_mask(super_cell, final_mask, params.verbose, num_levels,
                                            y, x, params.vert_dpi, params.horiz_dpi, N,
-                                           params.gamma, dot_grid, one_index);
+                                           params.gamma, dot_grid, one_index, mem);
         }
     }
     final_mask->bin_center = bin_center;
@@ -326,12 +351,13 @@ compare(const void * a, const void * b)
     double cost = val_a->value - val_b->value;
 
     /* If same value, use distance to center for decision */
-    if (cost == 0) {
-        /* Don't think these should ever be the same due to tiling effect */
-        return val_a->dist_to_center - val_b->dist_to_center;
-    } else {
-       return cost;
-    }
+    if (cost == 0)
+        cost = val_a->dist_to_center - val_b->dist_to_center;
+    if (cost == 0)
+        return 0;
+    if (cost < 0)
+        return -1;
+    return 1;
 }
 
 static int
@@ -380,7 +406,7 @@ htsc_mask_to_tos(htsc_dig_grid_t *final_mask)
     }
 #endif
         /* Sort */
-    qsort(values, height * width, sizeof(htsc_threshpoint_t), compare);
+    qsort(values, (size_t)height * width, sizeof(htsc_threshpoint_t), compare);
 #if RAW_SCREEN_DUMP
     EPRINTF(final_mask->memory, "Sorted\n");
     for (k = 0; k < count; k++) {
@@ -402,8 +428,8 @@ htsc_mask_to_tos(htsc_dig_grid_t *final_mask)
 }
 
 static void
-htsc_determine_cell_shape(double *x_out, double *y_out, double *v_out,
-                          double *u_out, double *N_out, htsc_param_t params, void *mem)
+htsc_determine_cell_shape(int *x_out, int *y_out, int *v_out,
+                          int *u_out, int *N_out, htsc_param_t params, void *mem)
 {
     double x = 0., y = 0., v = 0., u = 0., N = 0.;
     double frac, scaled_x;
@@ -482,7 +508,7 @@ htsc_determine_cell_shape(double *x_out, double *y_out, double *v_out,
         }
         x = x_use;
     }
-    if (ratio >= 1 && ratio!=0) {
+    if (ratio >= 1) {
         /* The minimal step is in y */
         if (params.verbose > 0) {
             PRINTF(mem, "x\ty\tu\tv\tAngle\tLPI\tLevels\n");
@@ -660,11 +686,11 @@ htsc_determine_cell_shape(double *x_out, double *y_out, double *v_out,
             x = x_use;
         }
     }
-    *x_out = x;
-    *y_out = y;
-    *v_out = v;
-    *u_out = u;
-    *N_out = N;
+    *x_out = (int)x;
+    *y_out = (int)y;
+    *v_out = (int)v;
+    *u_out = (int)u;
+    *N_out = (int)N;
 }
 
 static void
@@ -797,7 +823,7 @@ htsc_create_dot_mask(htsc_dig_grid_t *dot_grid, int x, int y, int u, int v,
     int k,j,val_min,index_x,index_y;
     double slope1, slope2;
     int t1,t2,t3,t4,point_in;
-    float b3, b4;
+    double b3, b4;
     htsc_point_t test_point;
 
     if (screen_angle != 0) {
@@ -807,10 +833,10 @@ htsc_create_dot_mask(htsc_dig_grid_t *dot_grid, int x, int y, int u, int v,
         dot_grid->height = abs(val_min) + y;
         dot_grid->width = x + u;
         dot_grid->data =
-            (int *) ALLOC(dot_grid->memory, dot_grid->height * dot_grid->width * sizeof(int));
+	  (int *) ALLOC(dot_grid->memory, (size_t)dot_grid->height * dot_grid->width * sizeof(int));
         if (dot_grid->data == NULL)
             return -1;
-        memset(dot_grid->data,0,dot_grid->height * dot_grid->width * sizeof(int));
+        memset(dot_grid->data,0,(size_t)dot_grid->height * dot_grid->width * sizeof(int));
         index_x = 0;
         for (k = 0; k < (x+u); k++) {
             index_y=0;
@@ -838,10 +864,10 @@ htsc_create_dot_mask(htsc_dig_grid_t *dot_grid, int x, int y, int u, int v,
         /* All points are valid */
         dot_grid->height = y;
         dot_grid->width = u;
-        dot_grid->data = (int *) ALLOC(dot_grid->memory, y * u * sizeof(int));
+        dot_grid->data = (int *) ALLOC(dot_grid->memory, (size_t)y * u * sizeof(int));
         if (dot_grid->data == NULL)
             return -1;
-        memset(dot_grid->data, -1, y * u * sizeof(int));
+        memset(dot_grid->data, -1, (size_t)y * u * sizeof(int));
         val_min = 0;
     }
     return 0;
@@ -1041,8 +1067,8 @@ htsc_allocate_supercell(htsc_dig_grid_t *super_cell, int x, int y, int u,
         min_vert_number = *H;
     }
 
-    a = ceil((float) target_size / (float) lcm_value);
-    b = ceil((float) target_size / (float) min_vert_number);
+    a = (int)ceil((double) target_size / (double) lcm_value);
+    b = (int)ceil((double) target_size / (double) min_vert_number);
 
     /* super_cell Size is  b*min_vert_number by a*lcm_value
        create the large cell */
@@ -1057,10 +1083,10 @@ htsc_allocate_supercell(htsc_dig_grid_t *super_cell, int x, int y, int u,
     super_cell->height = super_size_y;
     super_cell->width = super_size_x;
     super_cell->data =
-        (int *) ALLOC(dot_grid.memory, super_size_x * super_size_y * sizeof(int));
+        (int *) ALLOC(dot_grid.memory, (size_t)super_size_x * super_size_y * sizeof(int));
     if (super_cell->data == NULL)
         return -1;
-    memset(super_cell->data, 0, super_size_x * super_size_y * sizeof(int));
+    memset(super_cell->data, 0, (size_t)super_size_x * super_size_y * sizeof(int));
     return 0;
 }
 
@@ -1135,25 +1161,26 @@ htsc_tile_supercell(htsc_dig_grid_t *super_cell, htsc_dig_grid_t *dot_grid,
 
 /* Create 2d gaussian filter that varies with respect to coordinate
    spatial resolution */
-void
-create_2d_gauss_filter(float *filter, int x_size, int y_size,
-                       float stdvalx, float stdvaly)
+int
+create_2d_gauss_filter(double *filter, int x_size, int y_size,
+    double stdvalx, double stdvaly, gs_memory_t *mem)
 {
     int x_half_size  = (x_size-1)/2;
     int y_half_size = (y_size-1)/2;
     int k,j;
-    float arg, val;
+    double arg, val;
     double sum = 0;
-    float max_val = 0;
+    double max_val = 0;
     int total_size = x_size * y_size;
     int index_x, index_y;
+    int code = 0;
 
     for (j = -y_half_size; j < y_half_size+1; j++) {
         index_y = j + y_half_size;
         for (k = -x_half_size; k < x_half_size+1; k++) {
             arg   = -(k  * k / (stdvalx * stdvalx) +
                       j * j / (stdvaly * stdvaly) ) /2;
-            val = (float) exp(arg);
+            val = exp(arg);
             sum += val;
             if (val > max_val) max_val = val;
             index_x = k + x_half_size;
@@ -1164,16 +1191,17 @@ create_2d_gauss_filter(float *filter, int x_size, int y_size,
         filter[j]/=sum;
     }
 #if RAW_SCREEN_DUMP
-    htsc_dump_float_image(filter, y_size, x_size, max_val/sum, "guass_filt");
+    code = htsc_dump_float_image(filter, y_size, x_size, max_val/sum, "guass_filt", mem);
 #endif
+    return code;
 }
 
 /* 2-D convolution (or correlation) with periodic boundary condition */
 static void
 htsc_apply_filter(byte *screen_matrix, int num_cols_sc,
-                  int num_rows_sc, float *filter, int num_cols_filt,
-                  int num_rows_filt, float *screen_blur,
-                  float *max_val, htsc_point_t *max_pos, float *min_val,
+                  int num_rows_sc, double *filter, int num_cols_filt,
+                  int num_rows_filt, double *screen_blur,
+    double *max_val, htsc_point_t *max_pos, double *min_val,
                   htsc_point_t *min_pos)
 {
     int k,j,kk,jj;
@@ -1181,8 +1209,8 @@ htsc_apply_filter(byte *screen_matrix, int num_cols_sc,
     int half_cols_filt = (num_cols_filt-1)/2;
     int half_rows_filt = (num_rows_filt-1)/2;
     int j_circ,k_circ;
-    float fmax_val = -1;
-    float fmin_val = 100000000;
+    double fmax_val = -1;
+    double fmin_val = 100000000;
     htsc_point_t fmax_pos = { 0.0, 0.0 }, fmin_pos = { 0.0, 0.0 };
 
     for (j = 0; j < num_rows_sc; j++) {
@@ -1247,28 +1275,32 @@ htsc_add_dots(byte *screen_matrix, int num_cols, int num_rows,
     double sigma_y = vert_dpi / lpi_act;
     double sigma_x = sigma_y * xscale;
     int sizefiltx, sizefilty;
-    float *filter;
-    float *screen_blur;
+    double *filter;
+    double *screen_blur;
     int white_pos;
-    float max_val, min_val;
+    double max_val, min_val;
     htsc_point_t max_pos, min_pos;
     int k,j;
     int dist, curr_dist;
     htsc_dither_pos_t curr_position;
+    int code;
 
     sizefiltx = ROUND(sigma_x * 4);
     sizefilty = ROUND(sigma_y * 4);
-    if ( ((float) sizefiltx / 2.0) == (sizefiltx >> 1)) {
+    if ( ((double) sizefiltx / 2.0) == (sizefiltx >> 1)) {
         sizefiltx += 1;
     }
-    if ( ((float) sizefilty / 2.0) == (sizefilty >> 1)) {
+    if ( ((double) sizefilty / 2.0) == (sizefilty >> 1)) {
         sizefilty += 1;
     }
-    filter = (float*) ALLOC(mem, sizeof(float) * sizefilty * sizefiltx);
+    filter = (double*) ALLOC(mem, sizeof(double) * sizefilty * sizefiltx);
     if (filter == NULL)
         return -1;
-    create_2d_gauss_filter(filter, sizefiltx, sizefilty, sizefiltx, sizefilty);
-    screen_blur = (float*)ALLOC(mem, sizeof(float) * num_cols * num_rows);
+    code = create_2d_gauss_filter(filter, sizefiltx, sizefilty, (double)sizefiltx, (double)sizefilty, mem);
+    if (code < 0)
+        return code;
+
+    screen_blur = (double*)ALLOC(mem, sizeof(double) * num_cols * num_rows);
     if (screen_blur == NULL) {
         FREE(mem, filter);
         return -1;
@@ -1284,8 +1316,8 @@ htsc_add_dots(byte *screen_matrix, int num_cols, int num_rows,
         white_pos = 0;
         dist = (num_cols) * (num_cols) + (num_rows) * (num_rows);
         for (k = 0; k < num_dots; k++) {
-            curr_dist = (pos_y[k] - min_pos.y) * (pos_y[k] - min_pos.y) +
-                        (pos_x[k] - min_pos.x) * (pos_x[k] - min_pos.x);
+            curr_dist = (pos_y[k] - (int)min_pos.y) * (pos_y[k] - (int)min_pos.y) +
+                        (pos_x[k] - (int)min_pos.x) * (pos_x[k] - (int)min_pos.x);
             if (curr_dist < dist &&
                 screen_matrix[pos_x[k] + num_cols * pos_y[k]] == 0) {
                 white_pos = k;
@@ -1317,29 +1349,33 @@ htsc_init_dot_position(byte *screen_matrix, int num_cols, int num_rows,
     double sigma_y = vert_dpi / lpi_act;
     double sigma_x = sigma_y * xscale;
     int sizefiltx, sizefilty;
-    float *filter;
+    double *filter;
     bool done = false;
-    float *screen_blur;
+    double *screen_blur;
     int white_pos, black_pos;
-    float max_val, min_val;
+    double max_val, min_val;
     htsc_point_t max_pos, min_pos;
     int k;
     int dist, curr_dist;
     bool found_it;
+    int code = 0;
 
     sizefiltx = ROUND(sigma_x * 4);
     sizefilty = ROUND(sigma_y * 4);
-    if ( ((float) sizefiltx / 2.0) == (sizefiltx >> 1)) {
+    if ( ((double) sizefiltx / 2.0) == (sizefiltx >> 1)) {
         sizefiltx += 1;
     }
-    if ( ((float) sizefilty / 2.0) == (sizefilty >> 1)) {
+    if ( ((double) sizefilty / 2.0) == (sizefilty >> 1)) {
         sizefilty += 1;
     }
-    filter = (float*) ALLOC(mem, sizeof(float) * sizefilty * sizefiltx);
+    filter = (double*) ALLOC(mem, sizeof(double) * sizefilty * sizefiltx);
     if (filter == NULL)
         return -1;
-    create_2d_gauss_filter(filter, sizefiltx, sizefilty, sizefiltx, sizefilty);
-    screen_blur = (float*) ALLOC(mem, sizeof(float) * num_cols * num_rows);
+    code = create_2d_gauss_filter(filter, sizefiltx, sizefilty, (double)sizefiltx, (double)sizefilty, mem);
+    if (code < 0)
+        return code;
+
+    screen_blur = (double*) ALLOC(mem, sizeof(double) * num_cols * num_rows);
     if (screen_blur == NULL) {
         FREE(mem, filter);
         return -1;
@@ -1350,15 +1386,17 @@ htsc_init_dot_position(byte *screen_matrix, int num_cols, int num_rows,
         htsc_apply_filter(screen_matrix, num_cols, num_rows, filter, sizefiltx,
                   sizefilty, screen_blur, &max_val, &max_pos, &min_val, &min_pos);
 #if RAW_SCREEN_DUMP
-        htsc_dump_float_image(screen_blur, num_cols, num_rows, max_val, "blur_one");
+        code = htsc_dump_float_image(screen_blur, num_cols, num_rows, max_val, "blur_one", mem);
+        if (code < 0)
+            return code;
 #endif
     /* Find the closest on dot to the max position */
         black_pos = 0;
-        dist = (pos_y[0] - max_pos.y) * (pos_y[0] - max_pos.y) +
-               (pos_x[0] - max_pos.x) * (pos_x[0] - max_pos.x);
+        dist = (pos_y[0] - (int)max_pos.y) * (pos_y[0] - (int)max_pos.y) +
+               (pos_x[0] - (int)max_pos.x) * (pos_x[0] - (int)max_pos.x);
         for ( k = 1; k < num_dots; k++) {
-            curr_dist = (pos_y[k] - max_pos.y) * (pos_y[k] - max_pos.y) +
-                        (pos_x[k] - max_pos.x) * (pos_x[k] - max_pos.x);
+            curr_dist = (pos_y[k] - (int)max_pos.y) * (pos_y[k] - (int)max_pos.y) +
+                        (pos_x[k] - (int)max_pos.x) * (pos_x[k] - (int)max_pos.x);
             if (curr_dist < dist &&
                 screen_matrix[pos_x[k] + num_cols * pos_y[k]] == 1) {
                 black_pos = k;
@@ -1372,11 +1410,11 @@ htsc_init_dot_position(byte *screen_matrix, int num_cols, int num_rows,
                   sizefilty, screen_blur, &max_val, &max_pos, &min_val, &min_pos);
         /* Find the closest OFF dot to the min position. */
         white_pos = 0;
-        dist = (pos_y[0] - min_pos.y) * (pos_y[0] - min_pos.y) +
-               (pos_x[0] - min_pos.x) * (pos_x[0] - min_pos.x);
+        dist = (pos_y[0] - (int)min_pos.y) * (pos_y[0] - (int)min_pos.y) +
+               (pos_x[0] - (int)min_pos.x) * (pos_x[0] - (int)min_pos.x);
         for ( k = 1; k < num_dots; k++) {
-            curr_dist = (pos_y[k] - min_pos.y) * (pos_y[k] - min_pos.y) +
-                        (pos_x[k] - min_pos.x) * (pos_x[k] - min_pos.x);
+            curr_dist = (pos_y[k] - (int)min_pos.y) * (pos_y[k] - (int)min_pos.y) +
+                        (pos_x[k] - (int)min_pos.x) * (pos_x[k] - (int)min_pos.x);
             if (curr_dist < dist &&
                 screen_matrix[pos_x[k] + num_cols * pos_y[k]] == 0) {
                 white_pos = k;
@@ -1424,7 +1462,8 @@ static int
 htsc_create_dither_mask(htsc_dig_grid_t super_cell, htsc_dig_grid_t *final_mask,
                           int verbose, int num_levels, int y, int x, double vert_dpi,
                           double horiz_dpi, int N, double gamma,
-                          htsc_dig_grid_t dot_grid, htsc_point_t dot_grid_one_index)
+                          htsc_dig_grid_t dot_grid, htsc_point_t dot_grid_one_index,
+                          gs_memory_t *mem)
 {
     int *dot_levels = NULL;
     int *locate = NULL;
@@ -1519,7 +1558,7 @@ htsc_create_dither_mask(htsc_dig_grid_t super_cell, htsc_dig_grid_t *final_mask,
             code = -1;
             goto out;
         }
-        percent = 1.0 / ((float)num_levels + 1.0);
+        percent = 1.0 / ((double)num_levels + 1.0);
         prev_dot_level = 0;
         for (k = 0; k < num_levels; k++) {
             perc_val = (k + 1) * percent;
@@ -1561,8 +1600,10 @@ htsc_create_dither_mask(htsc_dig_grid_t super_cell, htsc_dig_grid_t *final_mask,
             }
         }
 #if RAW_SCREEN_DUMP
-        htsc_dump_byte_image(screen_matrix, width_supercell, height_supercell,
-                             1, "screen0_init");
+        code  = htsc_dump_byte_image(screen_matrix, width_supercell, height_supercell,
+                             1, "screen0_init", mem);
+        if (code < 0)
+            goto out;
 #endif
         /* Rearrange these dots into a pleasing pattern, but only if there is
         * more than 1.  Otherwise there are none to move */
@@ -1573,8 +1614,10 @@ htsc_create_dither_mask(htsc_dig_grid_t super_cell, htsc_dig_grid_t *final_mask,
         if (code < 0)
             goto out;
 #if RAW_SCREEN_DUMP
-        htsc_dump_byte_image(screen_matrix, width_supercell, height_supercell,
-                             1, "screen0_arrange");
+        code  = htsc_dump_byte_image(screen_matrix, width_supercell, height_supercell,
+                             1, "screen0_arrange", mem);
+        if (code < 0)
+            goto out;
 #endif
         /* Now we want to introduce more dots at each level */
         for (k = 1; k < num_levels; k++) {
@@ -1588,8 +1631,10 @@ htsc_create_dither_mask(htsc_dig_grid_t super_cell, htsc_dig_grid_t *final_mask,
             {
             char str_name[30];
             snprintf(str_name, 30, "screen%d_arrange",k);
-            htsc_dump_byte_image(screen_matrix, width_supercell, height_supercell,
-                                 1, str_name);
+            code = htsc_dump_byte_image(screen_matrix, width_supercell, height_supercell,
+                                 1, str_name, mem);
+            if (code < 0)
+                goto out;
             }
 #endif
         }
@@ -1615,7 +1660,7 @@ htsc_create_dither_mask(htsc_dig_grid_t super_cell, htsc_dig_grid_t *final_mask,
             goto out;
         }
         for (k = 0; k < N; k++) {
-            thresholds[N-1-k] = (k + 1) * step_size - (step_size / 2);
+            thresholds[N-1-k] = (int)((k + 1) * step_size - (step_size / 2));
         }
         mag_offset =
             (double) (thresholds[0]-thresholds[1]) / (double) (num_levels+1);
@@ -1688,9 +1733,9 @@ htsc_create_dither_mask(htsc_dig_grid_t super_cell, htsc_dig_grid_t *final_mask,
 
                         /* In case we have modulo of a negative number */
                         if (k_index < 0) k_index = k_index + width_supercell;
-                        threshold_value = thresholds[val-1] +
-                                          mag_offset * dot_level_sort[h];
-                        if (threshold_value > MAXVAL) threshold_value = MAXVAL;
+                        threshold_value = (int)(thresholds[val-1] +
+                                                mag_offset * dot_level_sort[h]);
+                        if (threshold_value > MAXVAL) threshold_value = (int)MAXVAL;
                         if (threshold_value < 0) threshold_value = 0;
                         htsc_setpoint(final_mask,k_index,j_index,threshold_value);
                     }
@@ -1732,13 +1777,13 @@ htsc_create_holladay_mask(htsc_dig_grid_t super_cell, int H, int L,
 
     final_mask->height = H;
     final_mask->width = L;
-    final_mask->data = (int *) ALLOC(final_mask->memory, H * L * sizeof(int));
+    final_mask->data = (int *) ALLOC(final_mask->memory, (size_t)H * L * sizeof(int));
     if (final_mask->data == NULL) {
         code = -1;
         goto out;
     }
 
-    thresholds = (double *) ALLOC(final_mask->memory, H * L * sizeof(double));
+    thresholds = (double *) ALLOC(final_mask->memory, (size_t)H * L * sizeof(double));
     if (final_mask->data == NULL) {
         code = -1;
         goto out;
@@ -1752,7 +1797,7 @@ htsc_create_holladay_mask(htsc_dig_grid_t super_cell, int H, int L,
          thresholds[number_points - k - 1] =
                                     ROUND(temp * MAXVAL * white_scale + 1);
     }
-    memset(final_mask->data, 0, H * L * sizeof(int));
+    memset(final_mask->data, 0, (size_t)H * L * sizeof(int));
 
     for (j = 0; j < H; j++) {
         for (k = 0; k < L; k++) {
@@ -1782,13 +1827,14 @@ htsc_create_nondithered_mask(htsc_dig_grid_t super_cell, int H, int L,
 
     final_mask->height = super_cell.height;
     final_mask->width = super_cell.width;
-    final_mask->data = (int *) ALLOC(final_mask->memory, super_cell.height * super_cell.width *
+    final_mask->data = (int *) ALLOC(final_mask->memory,
+				     (size_t)super_cell.height * super_cell.width *
                                       sizeof(int));
     if (final_mask->data == NULL) {
         code = -1;
         goto out;
     }
-    thresholds = (double *) ALLOC(final_mask->memory, H * L * sizeof(double));
+    thresholds = (double *) ALLOC(final_mask->memory, (size_t)H * L * sizeof(double));
     if (thresholds == NULL) {
         code = -1;
         goto out;
@@ -1802,7 +1848,7 @@ htsc_create_nondithered_mask(htsc_dig_grid_t super_cell, int H, int L,
          thresholds[number_points - k - 1] =
                                     ROUND(temp * MAXVAL * white_scale + 1);
     }
-    memset(final_mask->data, 0, super_cell.height * super_cell.width *
+    memset(final_mask->data, 0, (size_t)super_cell.height * super_cell.width *
                                 sizeof(int));
     for (j = 0; j < super_cell.height; j++) {
         for (k = 0; k < super_cell.width; k++) {
@@ -1924,13 +1970,161 @@ htsc_spot_value(spottype_t spot_type, double x, double y)
     }
 }
 
-#if RAW_SCREEN_DUMP
-void
-static
-htsc_dump_screen(htsc_dig_grid_t *dig_grid, char filename[])
+#if FINAL_SCREEN_DUMP
+
+/* Save turn on order list, Assume that htsc_gen_ordered has already converted to TOS */
+static int
+htsc_save_tos(htsc_dig_grid_t *final_mask, gs_memory_t *mem)
+{
+    int width = final_mask->width;
+    int height = final_mask->height;
+    int *buff_ptr;
+    _FILE *fid;
+    int k = 0;
+    int count = height * width;
+
+    fid = FOPEN(mem, "turn_on_seq.out", "w");
+    if (fid == NULL)
+        return -1;
+
+    FPRINTF2(fid, "# W=%d H=%d\n", width, height);
+
+    /* Write out */
+    buff_ptr = final_mask->data;
+    for (k = 0; k < count; k++) {
+        FPRINTF2(fid, "%d\t%d\n", *buff_ptr++, *buff_ptr++);
+    }
+    FCLOSE(fid);
+    return 0;
+}
+
+int
+htsc_save_screen(htsc_dig_grid_t *final_mask, bool use_holladay_grid, int S,
+    htsc_param_t params, gs_memory_t *mem)
 {
     char full_file_name[FULL_FILE_NAME_LENGTH];
-    FILE *fid;
+    _FILE *fid;
+    int x, y, code = 0;
+    int *buff_ptr = final_mask->data;
+    int width = final_mask->width;
+    int height = final_mask->height;
+    byte data;
+    unsigned short data_short;
+    output_format_type output_format = params.output_format;
+    char *output_extension = (output_format == OUTPUT_PS) ? "ps" :
+        ((output_format == OUTPUT_PPM) ? "ppm" :
+            ((output_format == OUTPUT_RAW16 ? "16.raw" : "raw")));
+
+    if (output_format == OUTPUT_TOS) {
+        /* We need to figure out the turn-on sequence from the threshold
+           array */
+        code = htsc_save_tos(final_mask, mem);
+    } else {
+        if (use_holladay_grid) {
+            snprintf(full_file_name, FULL_FILE_NAME_LENGTH, "Screen_Holladay_Shift%d_%dx%d.%s", S, width,
+                height, output_extension);
+        } else {
+            snprintf(full_file_name, FULL_FILE_NAME_LENGTH, "Screen_Dithered_%dx%d.%s", width, height,
+                output_extension);
+        }
+        fid = FOPEN(mem, full_file_name, "wb");
+        if (fid == NULL)
+            return -1;
+
+        if (output_format == OUTPUT_PPM)
+            FPRINTF6(fid, "P5\n"
+                "# Halftone threshold array, %s, [%d, %d], S=%d\n"
+                "%d %d\n"
+                "255\n",
+                use_holladay_grid ? "Holladay_Shift" : "Dithered", width, height,
+                S, width, height);
+        if (output_format != OUTPUT_PS) {
+            /* Both BIN and PPM format write the same binary data */
+            if (output_format == OUTPUT_RAW || output_format == OUTPUT_PPM) {
+                for (y = 0; y < height; y++) {
+                    for (x = 0; x < width; x++) {
+                        data_short = (unsigned short)(*buff_ptr & 0xffff);
+                        data_short = ROUND((double)data_short * 255.0 / MAXVAL);
+                        if (data_short > 255) data_short = 255;
+                        data = (byte)(data_short & 0xff);
+                        FWRITE(&data, sizeof(byte), 1, fid);
+                        buff_ptr++;
+                    }
+                }
+            } else {	/* raw16 data */
+                for (y = 0; y < height; y++) {
+                    for (x = 0; x < width; x++) {
+                        data_short = (unsigned short)(*buff_ptr & 0xffff);
+                        FWRITE(&data_short, sizeof(short), 1, fid);
+                        buff_ptr++;
+                    }
+                }
+            }
+        } else {	/* ps output format */
+            if (params.targ_quant <= 256) {
+                /* Output PS HalftoneType 3 dictionary */
+                FPRINTF2(fid, "%%!PS\n"
+                    "<< /HalftoneType 3\n"
+                    "   /Width  %d\n"
+                    "   /Height %d\n"
+                    "   /Thresholds <\n",
+                    width, height);
+
+                for (y = 0; y < height; y++) {
+                    for (x = 0; x < width; x++) {
+                        data_short = (unsigned short)(*buff_ptr & 0xffff);
+                        data_short = ROUND((double)data_short * 255.0 / MAXVAL);
+                        if (data_short > 255) data_short = 255;
+                        data = (byte)(data_short & 0xff);
+                        FPRINTF1(fid, "%02x", data);
+                        buff_ptr++;
+                        if ((x & 0x1f) == 0x1f && (x != (width - 1)))
+                            FPRINTF(fid, "\n");
+                    }
+                    FPRINTF(fid, "\n");
+                }
+                FPRINTF(fid, "   >\n"
+                    ">>\n"
+                );
+            } else {
+                /* Output PS HalftoneType 16 dictionary. Single array. */
+                FPRINTF(fid, "%%!PS\n"
+                    "%% Create a 'filter' from local hex data\n"
+                    "{ currentfile /ASCIIHexDecode filter /ReusableStreamDecode filter } exec\n");
+                /* hex data follows, 'file' object will be left on stack */
+                for (y = 0; y < height; y++) {
+                    for (x = 0; x < width; x++) {
+                        data_short = (unsigned short)(*buff_ptr & 0xffff);
+                        FPRINTF1(fid, "%04x", data_short);
+                        buff_ptr++;
+                        if ((x & 0x1f) == 0x1f && (x != (width - 1)))
+                            FPRINTF(fid, "\n");
+                    }
+                    FPRINTF(fid, "\n");	/* end of one row */
+                }
+                FPRINTF(fid, ">\n");	/* ASCIIHexDecode EOF */
+                FPRINTF2(fid,
+                    "<< /Thresholds 2 index    %% file object for the 16-bit data\n"
+                    "   /HalftoneType 16\n"
+                    "   /Width  %d\n"
+                    "   /Height %d\n"
+                    ">>\n"
+                    "exch pop     %% discard ResuableStreamDecode file leaving the Halftone dict.\n",
+                    width, height);
+            }
+        }
+        FCLOSE(fid);
+    }
+    return code;
+}
+#endif
+
+#if RAW_SCREEN_DUMP
+static int
+htsc_dump_screen(htsc_dig_grid_t *dig_grid, char filename[], gs_memory_t *mem)
+{
+    char full_file_name[FULL_FILE_NAME_LENGTH];
+    _FILE *fid;
     int x,y;
     int *buff_ptr = dig_grid->data;
     int width = dig_grid->width;
@@ -1938,7 +2132,9 @@ htsc_dump_screen(htsc_dig_grid_t *dig_grid, char filename[])
     byte data[3];
 
     snprintf(full_file_name, FULL_FILE_NAME_LENGTH, "Screen_%s_%dx%dx3.raw",filename,width,height);
-    fid = fopen(full_file_name,"wb");
+    fid = FOPEN(mem, full_file_name,"wb");
+    if (fid == NULL)
+        return -1;
 
     for (y = 0; y < height; y++) {
         for ( x = 0; x < width; x++ ) {
@@ -1955,58 +2151,67 @@ htsc_dump_screen(htsc_dig_grid_t *dig_grid, char filename[])
                 data[1] = *buff_ptr;
                 data[2] = *buff_ptr;
             }
-            fwrite(data,sizeof(unsigned char),3,fid);
+            FWRITE(data,sizeof(unsigned char),3,fid);
             buff_ptr++;
         }
     }
-    fclose(fid);
+    FCLOSE(fid);
+    return 0;
 }
 
-static void
-htsc_dump_float_image(float *image, int height, int width, float max_val,
-                      char filename[])
+static int
+htsc_dump_float_image(double *image, int height, int width, double max_val,
+                      char filename[], gs_memory_t *mem)
 {
     char full_file_name[FULL_FILE_NAME_LENGTH];
-    FILE *fid;
+    _FILE *fid;
     int x,y;
     int data;
     byte data_byte;
 
     snprintf(full_file_name, FULL_FILE_NAME_LENGTH, "Float_%s_%dx%d.raw",filename,width,height);
-    fid = fopen(full_file_name,"wb");
+    fid = FOPEN(mem, full_file_name,"wb");
+    if (fid == NULL)
+        return -1;
+
     for (y = 0; y < height; y++) {
         for ( x = 0; x < width; x++ ) {
             data = (255.0 * image[x + y * width] / max_val);
             if (data > 255) data = 255;
             if (data < 0) data = 0;
             data_byte = data;
-            fwrite(&data_byte,sizeof(byte),1,fid);
+            FWRITE(&data_byte,sizeof(byte),1,fid);
         }
     }
-    fclose(fid);
+    FCLOSE(fid);
+    return 0;
 }
 
-static void
-htsc_dump_byte_image(byte *image, int height, int width, float max_val,
-                      char filename[])
+static int
+htsc_dump_byte_image(byte *image, int height, int width, double max_val,
+                      char filename[], gs_memory_t *mem)
 {
     char full_file_name[FULL_FILE_NAME_LENGTH];
-    FILE *fid;
+    _FILE *fid;
     int x,y;
     int data;
     byte data_byte;
 
     snprintf(full_file_name, FULL_FILE_NAME_LENGTH, "ByteScaled_%s_%dx%d.raw",filename,width,height);
-    fid = fopen(full_file_name,"wb");
+    fid = FOPEN(mem, full_file_name,"wb");
+    if (fid == NULL)
+        return -1;
+
     for (y = 0; y < height; y++) {
         for ( x = 0; x < width; x++ ) {
             data = (255.0 * image[x + y * width] / max_val);
             if (data > 255) data = 255;
             if (data < 0) data = 0;
             data_byte = data;
-            fwrite(&data_byte,sizeof(byte),1,fid);
+            FWRITE(&data_byte,sizeof(byte),1,fid);
         }
     }
-    fclose(fid);
+    FCLOSE(fid);
+    return 0;
 }
 #endif
