@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2019 Artifex Software, Inc.
+/* Copyright (C) 2001-2023 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -32,10 +32,12 @@
 #include "gxcmap.h"
 #include "gxdevice.h"
 #include "gxdevmem.h"
+#include "gxdevsop.h"
 #include "gxiodev.h"
 #include "gxcspace.h"
 #include "gsicc_manage.h"
 #include "gscms.h"
+#include "gxgetbit.h"
 
 /* Include the extern for the device list. */
 extern_gs_lib_device_list();
@@ -70,6 +72,10 @@ gx_device_finalize(const gs_memory_t *cmem, void *vptr)
         rc_decrement(dev->PageList, "gx_device_finalize(PageList)");
         dev->PageList = 0;
     }
+    if (dev->NupControl) {
+        rc_decrement(dev->NupControl, "gx_device_finalize(NupControl)");
+        dev->NupControl = 0;
+    }
 
     if (dev->finalize)
         dev->finalize(dev);
@@ -156,17 +162,6 @@ gx_device_reloc_ptr(gx_device * dev, gc_state_t * gcst)
     return RELOC_OBJ(dev);	/* gcst implicit */
 }
 
-/* Set up the device procedures in the device structure. */
-/* Also copy old fields to new ones. */
-void
-gx_device_set_procs(gx_device * dev)
-{
-    if (dev->static_procs != 0) {	/* 0 if already populated */
-        dev->procs = *dev->static_procs;
-        dev->static_procs = 0;
-    }
-}
-
 /* Flush buffered output to the device */
 int
 gs_flushpage(gs_gstate * pgs)
@@ -201,7 +196,7 @@ gs_output_page(gs_gstate * pgs, int num_copies, int flush)
     ((code = gs_fill(pgs)) < 0))
     {
         gs_grestore(pgs);
-	return code;
+        return code;
     }
     code = gs_grestore(pgs);
     if (code < 0)
@@ -242,10 +237,27 @@ gs_copyscanlines(gx_device * dev, int start_y, byte * data, uint size,
     uint count = size / line_size;
     uint i;
     byte *dest = data;
+    gs_int_rect rect;
+    gs_get_bits_params_t params;
+
+    rect.p.x = 0;
+    rect.q.x = dev->width;
+    params.x_offset = 0;
+    params.raster = bitmap_raster(dev->width * dev->color_info.depth);
 
     for (i = 0; i < count; i++, dest += line_size) {
-        int code = (*dev_proc(dev, get_bits)) (dev, start_y + i, dest, NULL);
+        int code;
 
+        rect.p.y = start_y+i;
+        rect.q.y = start_y+i+1;
+
+        params.options = (GB_ALIGN_ANY |
+                          GB_RETURN_COPY |
+                          GB_OFFSET_0 |
+                          GB_RASTER_STANDARD | GB_PACKING_CHUNKY |
+                          GB_COLORS_NATIVE | GB_ALPHA_NONE);
+        params.data[0] = dest;
+        code = (*dev_proc(dev, get_bits_rectangle))(dev, &rect, &params);
         if (code < 0) {
             /* Might just be an overrun. */
             if (start_y + i == dev->height)
@@ -351,7 +363,7 @@ gx_device_make_struct_type(gs_memory_struct_type_t *st,
 {
     if (dev->stype)
         *st = *dev->stype;
-    else if (dev_proc(dev, get_xfont_procs) == gx_forward_get_xfont_procs)
+    else if (dev_proc(dev, get_page_device) == gx_forward_get_page_device)
         *st = st_device_forward;
     else
         *st = st_device;
@@ -404,29 +416,29 @@ gs_copydevice2(gx_device ** pnew_dev, const gx_device * dev, bool keep_open,
         gs_free_object(mem->non_gc_memory, a_std, "gs_copydevice(stype)");
         return_error(gs_error_VMerror);
     }
-    gx_device_init(new_dev, dev, mem, false);
-    gx_device_set_procs(new_dev);
+    code = gx_device_init(new_dev, dev, mem, false);
     new_dev->stype = new_std;
     new_dev->stype_is_dynamic = new_std != std;
     /*
      * keep_open is very dangerous.  On the other hand, so is copydevice in
      * general, since it just copies the bits without any regard to pointers
-     * (including self-pointers) that they may contain.  We handle this by
-     * making the default finish_copydevice forbid copying of anything other
-     * than the device prototype.
+     * (including self-pointers) that they may contain.
      */
     new_dev->is_open = dev->is_open && keep_open;
-    fill_dev_proc(new_dev, finish_copydevice, gx_default_finish_copydevice);
-    code = dev_proc(new_dev, finish_copydevice)(new_dev, dev);
     if (code < 0) {
         gs_free_object(mem, new_dev, "gs_copydevice(device)");
 #if 0 /* gs_free_object above calls gx_device_finalize,
-         which closes the device and releaszes its stype, i.e. a_std. */
+         which closes the device and releases its stype, i.e. a_std. */
         if (a_std)
             gs_free_object(dev->memory->non_gc_memory, a_std, "gs_copydevice(stype)");
 #endif
         return code;
     }
+    /* We really want to be able to interrogate the device for capabilities
+     * and/or preferences right from when it is created, so set dev_spec_op
+     * now (if not already set).
+     */
+    fill_dev_proc(new_dev, dev_spec_op, gx_default_dev_spec_op);
     *pnew_dev = new_dev;
     return 0;
 }
@@ -609,15 +621,25 @@ gs_setdevice_no_init(gs_gstate * pgs, gx_device * dev)
 }
 
 /* Initialize a just-allocated device. */
-void
+int
 gx_device_init(gx_device * dev, const gx_device * proto, gs_memory_t * mem,
                bool internal)
 {
     memcpy(dev, proto, proto->params_size);
-    dev->memory = mem;
+    dev->initialize_device_procs = proto->initialize_device_procs;
+    if (dev->initialize_device_procs != NULL)
+        dev->initialize_device_procs(dev);
+    dev->memory = mem;		/* must precede initialize_device call so devices can use it */
+    if (dev->procs.initialize_device) {
+        int code = dev->procs.initialize_device(dev);
+        if (code < 0)
+            return code;
+    }
     dev->retained = !internal;
     rc_init(dev, mem, (internal ? 0 : 1));
     rc_increment(dev->icc_struct);
+
+    return 0;
 }
 
 void
@@ -625,11 +647,19 @@ gx_device_init_on_stack(gx_device * dev, const gx_device * proto,
                         gs_memory_t * mem)
 {
     memcpy(dev, proto, proto->params_size);
+    dev->initialize_device_procs = proto->initialize_device_procs;
+    dev->initialize_device_procs(dev);
+    if (dev->procs.initialize_device) {
+        /* A condition of devices inited on the stack is that they can
+         * never fail to initialize! */
+        (void)dev->procs.initialize_device(dev);
+    }
+    gx_device_fill_in_procs(dev);
     dev->memory = mem;
     dev->retained = 0;
     dev->pad = proto->pad;
     dev->log2_align_mod = proto->log2_align_mod;
-    dev->is_planar = proto->is_planar;
+    dev->num_planar_planes = proto->num_planar_planes;
     rc_init(dev, NULL, 0);
 }
 
@@ -638,8 +668,11 @@ void
 gs_make_null_device(gx_device_null *dev_null, gx_device *dev,
                     gs_memory_t * mem)
 {
-    gx_device_init((gx_device *)dev_null, (const gx_device *)&gs_null_device,
-                   mem, true);
+    /* Can never fail */
+    (void)gx_device_init((gx_device *)dev_null,
+                         (const gx_device *)&gs_null_device,
+                         mem, true);
+    gx_device_fill_in_procs((gx_device *)dev_null);
     gx_device_set_target((gx_device_forward *)dev_null, dev);
     if (dev) {
         /* The gx_device_copy_color_params() call below should
@@ -660,7 +693,6 @@ gs_make_null_device(gx_device_null *dev_null, gx_device *dev,
         set_dev_proc(dn, begin_transparency_mask, gx_default_begin_transparency_mask);
         set_dev_proc(dn, end_transparency_mask, gx_default_end_transparency_mask);
         set_dev_proc(dn, discard_transparency_layer, gx_default_discard_transparency_layer);
-        set_dev_proc(dn, pattern_manage, gx_default_pattern_manage);
         set_dev_proc(dn, push_transparency_state, gx_default_push_transparency_state);
         set_dev_proc(dn, pop_transparency_state, gx_default_pop_transparency_state);
         set_dev_proc(dn, put_image, gx_default_put_image);
@@ -674,8 +706,13 @@ gs_make_null_device(gx_device_null *dev_null, gx_device *dev,
 /* Is a null device ? */
 bool gs_is_null_device(gx_device *dev)
 {
+    gx_device ldev;
+
+    ldev.initialize_device_procs = gs_null_device.initialize_device_procs;
+    ldev.initialize_device_procs(&ldev);
+
     /* Assuming null_fill_path isn't used elswhere. */
-    return dev->procs.fill_path == gs_null_device.procs.fill_path;
+    return dev->procs.fill_path == ldev.procs.fill_path;
 }
 
 /* Mark a device as retained or not retained. */
@@ -776,13 +813,24 @@ gx_set_device_only(gs_gstate * pgs, gx_device * dev)
 uint
 gx_device_raster(const gx_device * dev, bool pad)
 {
-    ulong bits = (ulong) dev->width * dev->color_info.depth;
+    int depth = dev->color_info.depth;
+    ulong bits = (ulong) dev->width * depth;
     ulong raster;
     int l2align;
 
-    if (dev->is_planar)
-        bits /= dev->color_info.num_components;
+    if (dev->num_planar_planes) {
+        int num_components = dev->num_planar_planes;
+        /* bpc accounts for unused bits, e.g. depth==4, num_comp==3, or depth==8, num_comps==5 */
+        int bpc = depth / num_components;
 
+        /* depth can be <= num_components if planar and MEM_SET_PARAMS has changed it */
+        if (depth <= num_components || bpc >= 8) {
+            bits /= num_components;
+        } else {
+            /* depth is original depth, not the plane_depth since it is > num_components */
+            bits /= (depth / bpc);
+        }
+    }
     raster = (uint)((bits + 7) >> 3);
     if (!pad)
         return raster;
@@ -811,7 +859,7 @@ uint
 gx_device_raster_plane(const gx_device * dev, const gx_render_plane_t *render_plane)
 {
     ulong bpc = (render_plane && render_plane->index >= 0 ?
-        render_plane->depth : dev->color_info.depth/(dev->is_planar ? dev->color_info.num_components : 1));
+        render_plane->depth : dev->color_info.depth/(dev->num_planar_planes ? dev->num_planar_planes : 1));
     ulong bits = (ulong) dev->width * bpc;
     int l2align;
 
@@ -864,9 +912,21 @@ gx_device_set_hwsize_from_media(gx_device *dev)
     int rot = (dev->LeadingEdge & 1);
     double rot_media_x = rot ? dev->MediaSize[1] : dev->MediaSize[0];
     double rot_media_y = rot ? dev->MediaSize[0] : dev->MediaSize[1];
+    gx_device *parent = dev;
+    int hwsize[2];
 
-    dev->width = (int)(rot_media_x * dev->HWResolution[0] / 72.0 + 0.5);
-    dev->height = (int)(rot_media_y * dev->HWResolution[1] / 72.0 + 0.5);
+    /* Try the spec_op to give the device to control it */
+    hwsize[0] = (int)(rot_media_x * dev->HWResolution[0] / 72.0 + 0.5);
+    hwsize[1] = (int)(rot_media_y * dev->HWResolution[1] / 72.0 + 0.5);
+
+    while (parent->parent != NULL) {
+        parent = parent->parent;
+    }
+    if (dev_proc(parent, dev_spec_op)(parent, gxdso_set_HWSize, &hwsize, sizeof(hwsize)) <= 0) {
+        /* just do the default setting */
+        dev->width = hwsize[0];
+        dev->height = hwsize[1];
+    }
 }
 
 static void
@@ -1147,13 +1207,15 @@ int gx_device_delete_output_file(const gx_device * dev, const char *fname)
     const char *fmt;
     char *pfname = (char *)gs_alloc_bytes(dev->memory, gp_file_name_sizeof, "gx_device_delete_output_file(pfname)");
     int code;
+    size_t len;
 
     if (pfname == NULL) {
         code = gs_note_error(gs_error_VMerror);
-	goto done;
+        goto done;
     }
 
-    code = gx_parse_output_file_name(&parsed, &fmt, fname, strlen(fname),
+    len = strlen(fname);
+    code = gx_parse_output_file_name(&parsed, &fmt, fname, len,
                                          dev->memory);
     if (code < 0) {
         goto done;
@@ -1168,11 +1230,11 @@ int gx_device_delete_output_file(const gx_device * dev, const char *fname)
         while (*fmt != 'l' && *fmt != '%')
             --fmt;
         if (*fmt == 'l')
-            gs_sprintf(pfname, parsed.fname, count1);
+            gs_snprintf(pfname, gp_file_name_sizeof, parsed.fname, count1);
         else
-            gs_sprintf(pfname, parsed.fname, (int)count1);
+            gs_snprintf(pfname, gp_file_name_sizeof, parsed.fname, (int)count1);
     } else if (parsed.len && strchr(parsed.fname, '%'))	/* filename with "%%" but no "%nnd" */
-        gs_sprintf(pfname, parsed.fname);
+        gs_snprintf(pfname, gp_file_name_sizeof, parsed.fname);
     else
         pfname[0] = 0; /* 0 to use "fname", not "pfname" */
     if (pfname[0]) {
@@ -1180,7 +1242,7 @@ int gx_device_delete_output_file(const gx_device * dev, const char *fname)
         parsed.len = strlen(parsed.fname);
     }
     if (parsed.iodev)
-        code = parsed.iodev->procs.delete_file((gx_io_device *)(&parsed.iodev), (const char *)parsed.fname);
+        code = parsed.iodev->procs.delete_file((gx_io_device *)(parsed.iodev), (const char *)parsed.fname);
     else
         code = gs_note_error(gs_error_invalidfileaccess);
 
@@ -1209,7 +1271,7 @@ gx_device_open_output_file(const gx_device * dev, char *fname,
 
     if (pfname == NULL) {
         code = gs_note_error(gs_error_VMerror);
-	goto done;
+        goto done;
     }
 
     if (strlen(fname) == 0) {
@@ -1225,8 +1287,8 @@ gx_device_open_output_file(const gx_device * dev, char *fname,
     if (parsed.iodev && !strcmp(parsed.iodev->dname, "%stdout%")) {
         if (parsed.fname) {
             code = gs_note_error(gs_error_undefinedfilename);
-	    goto done;
-	}
+            goto done;
+        }
         *pfile = gp_file_FILE_alloc(dev->memory);
         if (*pfile == NULL) {
             code = gs_note_error(gs_error_VMerror);
@@ -1235,7 +1297,7 @@ gx_device_open_output_file(const gx_device * dev, char *fname,
         gp_file_FILE_set(*pfile, dev->memory->gs_lib_ctx->core->fstdout, noclose);
         /* Force stdout to binary. */
         code = gp_setmode_binary_impl(dev->memory->gs_lib_ctx->core->fstdout, true);
-	goto done;
+        goto done;
     } else if (parsed.iodev && !strcmp(parsed.iodev->dname, "%pipe%")) {
         positionable = false;
     }
@@ -1245,11 +1307,11 @@ gx_device_open_output_file(const gx_device * dev, char *fname,
         while (*fmt != 'l' && *fmt != '%')
             --fmt;
         if (*fmt == 'l')
-            gs_sprintf(pfname, parsed.fname, count1);
+            gs_snprintf(pfname, gp_file_name_sizeof, parsed.fname, count1);
         else
-            gs_sprintf(pfname, parsed.fname, (int)count1);
+            gs_snprintf(pfname, gp_file_name_sizeof, parsed.fname, (int)count1);
     } else if (parsed.len && strchr(parsed.fname, '%'))	/* filename with "%%" but no "%nnd" */
-        gs_sprintf(pfname, parsed.fname);
+        gs_snprintf(pfname, gp_file_name_sizeof, parsed.fname);
     else
         pfname[0] = 0; /* 0 to use "fname", not "pfname" */
     if (pfname[0]) {
@@ -1262,8 +1324,8 @@ gx_device_open_output_file(const gx_device * dev, char *fname,
 
         if (!parsed.fname) {
             code = gs_note_error(gs_error_undefinedfilename);
-	    goto done;
-	}
+            goto done;
+        }
         strcpy(fmode, gp_fmode_wb);
         if (positionable)
             strcat(fmode, "+");
@@ -1334,7 +1396,7 @@ bool gx_color_info_equal(const gx_device_color_info * p1, const gx_device_color_
         return false;
     if (p1->max_components != p2->max_components)
         return false;
-    if (p1->opmode != p2->opmode)
+    if (p1->opmsupported != p2->opmsupported)
         return false;
     if (p1->polarity != p2->polarity)
         return false;
@@ -1346,3 +1408,55 @@ bool gx_color_info_equal(const gx_device_color_info * p1, const gx_device_color_
         return false;
     return true;
 }
+
+int gx_callout(gx_device *dev, int id, int size, void *data)
+{
+    return gs_lib_ctx_callout(dev->memory, dev->dname,
+                              id, size, data);
+}
+
+/* compare two space_params, we can't do this with memcmp since there is padding in the structure */
+int
+gdev_space_params_cmp(const gdev_space_params sp1,
+                      const gdev_space_params sp2) {
+  if (sp1.MaxBitmap != sp2.MaxBitmap)
+    return(1);
+  if (sp1.BufferSpace != sp2.BufferSpace)
+    return(1);
+  if (sp1.band.BandWidth != sp2.band.BandWidth)
+    return(1);
+  if (sp1.band.BandHeight != sp2.band.BandHeight)
+    return(1);
+  if (sp1.band.BandBufferSpace != sp2.band.BandBufferSpace)
+    return(1);
+  if (sp1.band.tile_cache_size != sp2.band.tile_cache_size)
+    return(1);
+  if (sp1.params_are_read_only != sp2.params_are_read_only)
+    return(1);
+  if (sp1.banding_type != sp2.banding_type)
+    return(1);
+
+  return(0);
+}
+
+static void
+release_nts_lock(gx_device *dev)
+{
+    (void)gs_lib_ctx_nts_adjust(dev->memory, -1);
+}
+
+int gx_init_non_threadsafe_device(gx_device *dev)
+{
+    int code;
+
+    if (dev == NULL || dev->finalize != NULL)
+        return gs_error_unknownerror;
+
+    code = gs_lib_ctx_nts_adjust(dev->memory, 1);
+    if (code < 0)
+        return code;
+
+    dev->finalize = release_nts_lock;
+
+    return 0;
+}
