@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2019 Artifex Software, Inc.
+/* Copyright (C) 2001-2023 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 /* PDF 1.4 blending functions */
@@ -201,9 +201,9 @@ copy_extra_planes(byte *des_buf, pdf14_buf *des_info, byte *src_buf,
                   pdf14_buf *src_info, int width, int height)
 {
     /* alpha_g and shape do not copy */
-    des_buf += des_info->planestride * ((des_info->has_shape ? 1 : 0) +
+    des_buf += des_info->planestride * (size_t)((des_info->has_shape ? 1 : 0) +
                                         (des_info->has_alpha_g ? 1 : 0));
-    src_buf += src_info->planestride * ((src_info->has_shape ? 1 : 0) +
+    src_buf += src_info->planestride * (size_t)((src_info->has_shape ? 1 : 0) +
                                         (src_info->has_alpha_g ? 1 : 0));
     /* tags plane does copy */
     if (des_info->has_tags) {
@@ -226,6 +226,7 @@ pdf14_preserve_backdrop_cm(pdf14_buf *buf, cmm_profile_t *group_profile,
     int y0 = max(buf->rect.p.y, tos->rect.p.y);
     int y1 = min(buf->rect.q.y, tos->rect.q.y);
     bool deep = buf->deep;
+    int code;
 
     if (x0 < x1 && y0 < y1) {
         int width = x1 - x0;
@@ -241,7 +242,7 @@ pdf14_preserve_backdrop_cm(pdf14_buf *buf, cmm_profile_t *group_profile,
         rendering_params.graphics_type_tag = GS_IMAGE_TAG;
         rendering_params.override_icc = false;
         rendering_params.preserve_black = gsBKPRESNOTSPECIFIED;
-        rendering_params.rendering_intent = gsPERCEPTUAL;
+        rendering_params.rendering_intent = gsRELATIVECOLORIMETRIC;
         rendering_params.cmm = gsCMM_DEFAULT;
         /* Request the ICC link for the transform that we will need to use */
         icc_link = gsicc_get_link_profile(pgs, dev, tos_profile, group_profile,
@@ -260,20 +261,20 @@ pdf14_preserve_backdrop_cm(pdf14_buf *buf, cmm_profile_t *group_profile,
         } else {
             if (knockout_buff) {
                 buf_plane = buf->backdrop + ((x0 - buf->rect.p.x)<<deep) +
-                        (y0 - buf->rect.p.y) * buf->rowstride;
+                        (y0 - buf->rect.p.y) * (size_t)buf->rowstride;
                 tos_plane = tos->backdrop + ((x0 - tos->rect.p.x)<<deep) +
-                        (y0 - tos->rect.p.y) * tos->rowstride;
-                memset(buf->backdrop, 0, buf->n_chan * buf->planestride<<deep);
+                        (y0 - tos->rect.p.y) * (size_t)tos->rowstride;
+                memset(buf->backdrop, 0, buf->n_chan * ((size_t)buf->planestride)<<deep);
             } else {
                 buf_plane = buf->data + ((x0 - buf->rect.p.x)<<deep) +
-                        (y0 - buf->rect.p.y) * buf->rowstride;
+                        (y0 - buf->rect.p.y) * (size_t)buf->rowstride;
                 tos_plane = tos->data + ((x0 - tos->rect.p.x)<<deep) +
-                        (y0 - tos->rect.p.y) * tos->rowstride;
+                        (y0 - tos->rect.p.y) * (size_t)tos->rowstride;
                 /* First clear out everything. There are cases where the incoming buf
                    has a region outside the existing tos group.  Need to check if this
                    is getting clipped in which case we need to fix the allocation of
                    the buffer to be smaller */
-                memset(buf->data, 0, buf->n_planes * buf->planestride<<deep);
+                memset(buf->data, 0, buf->n_planes * ((size_t)buf->planestride)<<deep);
             }
             /* Set up the buffer descriptors. */
             gsicc_init_buffer(&input_buff_desc, tos_profile->num_comps, 1<<deep, false,
@@ -283,13 +284,15 @@ pdf14_preserve_backdrop_cm(pdf14_buf *buf, cmm_profile_t *group_profile,
                               false, true, buf->planestride, buf->rowstride, height,
                               width);
             /* Transform the data.  */
-            (icc_link->procs.map_buffer)(dev, icc_link, &input_buff_desc,
+            code = (icc_link->procs.map_buffer)(dev, icc_link, &input_buff_desc,
                                          &output_buff_desc, tos_plane, buf_plane);
             gsicc_release_link(icc_link);
+            if (code < 0)
+                return gs_throw(gs_error_unknownerror, "ICC transform failed.  Trans backdrop");
         }
         /* Copy the alpha data */
-        buf_plane += buf->planestride * (buf->n_chan - 1);
-        tos_plane += tos->planestride * (tos->n_chan - 1);
+        buf_plane += (buf->planestride) * (size_t)(buf->n_chan - 1);
+        tos_plane += (tos->planestride) * (size_t)(tos->n_chan - 1);
         copy_plane_part(buf_plane, buf->rowstride, tos_plane, tos->rowstride, width,
                         height, deep);
         buf_plane += buf->planestride;
@@ -301,7 +304,7 @@ pdf14_preserve_backdrop_cm(pdf14_buf *buf, cmm_profile_t *group_profile,
 #if RAW_DUMP
     if (x0 < x1 && y0 < y1) {
         byte *buf_plane = buf->data + ((x0 - buf->rect.p.x)<<deep) +
-            (y0 - buf->rect.p.y) * buf->rowstride;
+            (y0 - buf->rect.p.y) * (size_t)buf->rowstride;
         dump_raw_buffer(dev->memory, y1 - y0, x1 - x0, buf->n_planes, buf->planestride,
                         buf->rowstride, "BackDropInit_CM", buf_plane, deep);
         global_index++;
@@ -346,7 +349,7 @@ pdf14_preserve_backdrop(pdf14_buf *buf, pdf14_buf *tos, bool from_backdrop
             y0 > buf->rect.p.y || y1 < buf->rect.q.y) {
             /* FIXME: There is potential for more optimisation here,
              * but I don't know how often we hit this case. */
-            memset(buf_plane, 0, n_planes * buf->planestride);
+            memset(buf_plane, 0, n_planes * (size_t)buf->planestride);
         } else if (n_planes > tos->n_chan) {
             /* The next planes are alpha_g, shape, tags. We need to clear
              * alpha_g and shape, but don't need to clear the tag plane
@@ -355,11 +358,12 @@ pdf14_preserve_backdrop(pdf14_buf *buf, pdf14_buf *tos, bool from_backdrop
             if (!from_backdrop && n_planes > tag_plane_num)
                 n_planes = tag_plane_num;
             if (n_planes > tos->n_chan)
-                memset(buf->data + tos->n_chan * buf->planestride, 0, (n_planes - tos->n_chan) * buf->planestride);
+                memset(buf->data + tos->n_chan * (size_t)buf->planestride, 0,
+                       (n_planes - tos->n_chan) * (size_t)buf->planestride);
         }
-        buf_plane += (y0 - buf->rect.p.y) * buf->rowstride +
+        buf_plane += (y0 - buf->rect.p.y) * (size_t)buf->rowstride +
                      ((x0 - buf->rect.p.x)<<deep);
-        tos_plane += (y0 - tos->rect.p.y) * tos->rowstride +
+        tos_plane += (y0 - tos->rect.p.y) * (size_t)tos->rowstride +
                      ((x0 - tos->rect.p.x)<<deep);
         /* Color and alpha plane */
         for (i = 0; i < tos->n_chan; i++) {
@@ -376,7 +380,7 @@ pdf14_preserve_backdrop(pdf14_buf *buf, pdf14_buf *tos, bool from_backdrop
         byte *buf_plane = (from_backdrop ? buf->backdrop : buf->data);
         if (buf_plane != NULL) {
             buf_plane += ((x0 - buf->rect.p.x) << buf->deep) +
-                (y0 - buf->rect.p.y) * buf->rowstride;
+                (y0 - buf->rect.p.y) * (size_t)buf->rowstride;
             dump_raw_buffer(mem, y1 - y0, x1 - x0, buf->n_planes, buf->planestride,
                 buf->rowstride, "BackDropInit", buf_plane, buf->deep);
             global_index++;
@@ -490,7 +494,7 @@ pdf14_decode_color16(gx_device * dev, gx_color_index color, gx_color_value * out
 }
 
 void
-pdf14_gray_cs_to_cmyk_cm(gx_device * dev, frac gray, frac out[])
+pdf14_gray_cs_to_cmyk_cm(const gx_device * dev, frac gray, frac out[])
 {
     uchar num_comp = dev->color_info.num_components;
 
@@ -502,7 +506,7 @@ pdf14_gray_cs_to_cmyk_cm(gx_device * dev, frac gray, frac out[])
 
 /* These three must handle rgb + spot */
 void
-pdf14_gray_cs_to_rgbspot_cm(gx_device * dev, frac gray, frac out[])
+pdf14_gray_cs_to_rgbspot_cm(const gx_device * dev, frac gray, frac out[])
 {
     uchar num_comp = dev->color_info.num_components;
 
@@ -512,7 +516,7 @@ pdf14_gray_cs_to_rgbspot_cm(gx_device * dev, frac gray, frac out[])
 }
 
 void
-pdf14_rgb_cs_to_rgbspot_cm(gx_device * dev, const gs_gstate *pgs,
+pdf14_rgb_cs_to_rgbspot_cm(const gx_device * dev, const gs_gstate *pgs,
     frac r, frac g, frac b, frac out[])
 {
     uchar num_comp = dev->color_info.num_components;
@@ -525,7 +529,7 @@ pdf14_rgb_cs_to_rgbspot_cm(gx_device * dev, const gs_gstate *pgs,
 }
 
 void
-pdf14_cmyk_cs_to_rgbspot_cm(gx_device * dev, frac c, frac m, frac y, frac k, frac out[])
+pdf14_cmyk_cs_to_rgbspot_cm(const gx_device * dev, frac c, frac m, frac y, frac k, frac out[])
 {
     uchar num_comp = dev->color_info.num_components;
 
@@ -536,7 +540,7 @@ pdf14_cmyk_cs_to_rgbspot_cm(gx_device * dev, frac c, frac m, frac y, frac k, fra
 
 /* These three must handle gray + spot */
 void
-pdf14_gray_cs_to_grayspot_cm(gx_device * dev, frac gray, frac out[])
+pdf14_gray_cs_to_grayspot_cm(const gx_device * dev, frac gray, frac out[])
 {
     uchar num_comp = dev->color_info.num_components;
 
@@ -546,7 +550,7 @@ pdf14_gray_cs_to_grayspot_cm(gx_device * dev, frac gray, frac out[])
 }
 
 void
-pdf14_rgb_cs_to_grayspot_cm(gx_device * dev, const gs_gstate *pgs,
+pdf14_rgb_cs_to_grayspot_cm(const gx_device * dev, const gs_gstate *pgs,
     frac r, frac g, frac b, frac out[])
 {
     uchar num_comp = dev->color_info.num_components;
@@ -557,7 +561,7 @@ pdf14_rgb_cs_to_grayspot_cm(gx_device * dev, const gs_gstate *pgs,
 }
 
 void
-pdf14_cmyk_cs_to_grayspot_cm(gx_device * dev, frac c, frac m, frac y, frac k, frac out[])
+pdf14_cmyk_cs_to_grayspot_cm(const gx_device * dev, frac c, frac m, frac y, frac k, frac out[])
 {
     uchar num_comp = dev->color_info.num_components;
 
@@ -582,7 +586,7 @@ pdf14_cmyk_cs_to_grayspot_cm(gx_device * dev, frac c, frac m, frac y, frac k, fr
  * often they are { pop 0 }.
  */
 void
-pdf14_rgb_cs_to_cmyk_cm(gx_device * dev, const gs_gstate *pgs,
+pdf14_rgb_cs_to_cmyk_cm(const gx_device * dev, const gs_gstate *pgs,
                            frac r, frac g, frac b, frac out[])
 {
     uchar num_comp = dev->color_info.num_components;
@@ -603,7 +607,7 @@ pdf14_rgb_cs_to_cmyk_cm(gx_device * dev, const gs_gstate *pgs,
 }
 
 void
-pdf14_cmyk_cs_to_cmyk_cm(gx_device * dev, frac c, frac m, frac y, frac k, frac out[])
+pdf14_cmyk_cs_to_cmyk_cm(const gx_device * dev, frac c, frac m, frac y, frac k, frac out[])
 {
     uchar num_comp = dev->color_info.num_components;
 
@@ -620,14 +624,14 @@ pdf14_cmyk_cs_to_cmyk_cm(gx_device * dev, frac c, frac m, frac y, frac k, frac o
 static	int
 dump_planar_rgba(gs_memory_t *mem, const pdf14_buf *pbuf)
 {
-    int rowstride = pbuf->rowstride, planestride = pbuf->planestride;
+    size_t rowstride = pbuf->rowstride, planestride = pbuf->planestride;
     int rowbytes = width << 2;
     gs_int_rect rect = buf->rect;
     int x1 = min(pdev->width, rect.q.x);
     int y1 = min(pdev->height, rect.q.y);
     int width = x1 - rect.p.x;
     int height = y1 - rect.p.y;
-    byte *buf_ptr = buf->data + rect.p.y * buf->rowstride + rect.p.x;
+    byte *buf_ptr = buf->data + rect.p.y * (size_t)buf->rowstride + rect.p.x;
     byte *row = gs_malloc(mem, rowbytes, 1, "png raster buffer");
     png_struct *png_ptr =
     png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
@@ -678,7 +682,7 @@ dump_planar_rgba(gs_memory_t *mem, const pdf14_buf *pbuf)
     info_ptr->color_type = PNG_COLOR_TYPE_RGB_ALPHA;
 
     /* add comment */
-    gs_sprintf(software_text, "%s %d.%02d", gs_product,
+    gs_snprintf(software_text, sizeof(software_text), "%s %d.%02d", gs_product,
             (int)(gs_revision / 100), (int)(gs_revision % 100));
     text_png.compression = -1;	/* uncompressed */
     text_png.key = (char *)software_key;	/* not const, unfortunately */
@@ -701,8 +705,8 @@ dump_planar_rgba(gs_memory_t *mem, const pdf14_buf *pbuf)
         for (x = 0; x < width; ++x) {
             row[(x << 2)] = buf_ptr[x];
             row[(x << 2) + 1] = buf_ptr[x + planestride];
-            row[(x << 2) + 2] = buf_ptr[x + planestride * 2];
-            row[(x << 2) + 3] = buf_ptr[x + planestride * 3];
+            row[(x << 2) + 2] = buf_ptr[x + planestride * (size_t)2];
+            row[(x << 2) + 3] = buf_ptr[x + planestride * (size_t)3];
         }
         png_write_row(png_ptr, row);
         buf_ptr += rowstride;
@@ -725,7 +729,7 @@ void
 gx_build_blended_image_row(const byte *gs_restrict buf_ptr, int planestride,
                            int width, int num_comp, uint16_t bg, byte *gs_restrict linebuf)
 {
-    int inc = planestride * num_comp;
+    size_t inc = planestride * (size_t)num_comp;
 
     buf_ptr += inc - 1;
     for (; width > 0; width--) {
@@ -763,12 +767,12 @@ gx_build_blended_image_row16(const byte *gs_restrict buf_ptr_, int planestride,
                              int width, int num_comp, uint16_t bg, byte *gs_restrict linebuf)
 {
     const uint16_t *gs_restrict buf_ptr = (const uint16_t *)(const void *)buf_ptr_;
-    int inc;
+    size_t inc;
 
     /* Note that we read in in native endian and blend,
      * then store out in big endian. */
     planestride >>= 1; /* Array indexing, not byte indexing */
-    inc = planestride * num_comp;
+    inc = planestride * (size_t)num_comp;
     buf_ptr += inc - 1;
     for (; width > 0; width--) {
         /* composite RGBA (or CMYKA, etc.) pixel with over solid background */
@@ -818,18 +822,18 @@ gx_blend_image_buffer(byte *buf_ptr, int width, int height, int rowstride,
         position = y * rowstride;
         for (x = 0; x < width; x++) {
             /* composite RGBA (or CMYKA, etc.) pixel with over solid background */
-            a = buf_ptr[position + planestride * num_comp];
+            a = buf_ptr[position + planestride * (size_t)num_comp];
             if ((a + 1) & 0xfe) {
                 a ^= 0xff;
                 for (comp_num = 0; comp_num < num_comp; comp_num++) {
-                    comp  = buf_ptr[position + planestride * comp_num];
+                    comp  = buf_ptr[position + planestride * (size_t)comp_num];
                     tmp = ((bg - comp) * a) + 0x80;
                     comp += (tmp + (tmp >> 8)) >> 8;
-                    buf_ptr[position + planestride * comp_num] = comp;
+                    buf_ptr[position + planestride * (size_t)comp_num] = comp;
                 }
             } else if (a == 0) {
                 for (comp_num = 0; comp_num < num_comp; comp_num++) {
-                    buf_ptr[position + planestride * comp_num] = bg;
+                    buf_ptr[position + planestride * (size_t)comp_num] = bg;
                 }
             }
             position+=1;
@@ -839,7 +843,7 @@ gx_blend_image_buffer(byte *buf_ptr, int width, int height, int rowstride,
 
 void
 gx_blend_image_buffer16(byte *buf_ptr_, int width, int height, int rowstride,
-                        int planestride, int num_comp, uint16_t bg)
+                        int planestride, int num_comp, uint16_t bg, bool keep_native)
 {
     uint16_t *buf_ptr = (uint16_t *)(void *)buf_ptr_;
     int x, y;
@@ -861,31 +865,33 @@ gx_blend_image_buffer16(byte *buf_ptr_, int width, int height, int rowstride,
         position = y * rowstride;
         for (x = 0; x < width; x++) {
             /* composite RGBA (or CMYKA, etc.) pixel with over solid background */
-            a = buf_ptr[position + planestride * num_comp];
+            a = buf_ptr[position + planestride * (size_t)num_comp];
             if (a == 0) {
                 for (comp_num = 0; comp_num < num_comp; comp_num++) {
-                    buf_ptr[position + planestride * comp_num] = bebg;
+                    buf_ptr[position + planestride * (size_t)comp_num] = bebg;
                 }
             } else if (a == 0xffff) {
-                /* Convert from native -> big endian */
-                /* FIXME: Are compilers smart enough to spot that this is
-                 * a no-op on big endian hosts? */
-                for (comp_num = 0; comp_num < num_comp; comp_num++) {
-                    comp  = buf_ptr[position + planestride * comp_num];
-                    ((byte *)&buf_ptr[position + planestride * comp_num])[0] = comp>>8;
-                    ((byte *)&buf_ptr[position + planestride * comp_num])[1] = comp;
+#if ARCH_IS_BIG_ENDIAN
+#else
+                if (!keep_native) {
+                    for (comp_num = 0; comp_num < num_comp; comp_num++) {
+                        comp = buf_ptr[position + planestride * (size_t)comp_num];
+                        ((byte *)&buf_ptr[position + planestride * (size_t)comp_num])[0] = comp >> 8;
+                        ((byte *)&buf_ptr[position + planestride * (size_t)comp_num])[1] = comp;
+                    }
                 }
+#endif
             } else {
                 a ^= 0xffff;
                 a += a>>15; /* a is now 0 to 0x10000 */
                 a >>= 1; /* We can only use 15 bits as bg-comp has a sign bit we can't lose */
                 for (comp_num = 0; comp_num < num_comp; comp_num++) {
-                    comp  = buf_ptr[position + planestride * comp_num];
+                    comp  = buf_ptr[position + planestride * (size_t)comp_num];
                     tmp = (((int)bg - comp) * a) + 0x4000;
                     comp += (tmp >> 15); /* Errors in bit 16 upwards will be ignored */
                     /* Store as big endian */
-                    ((byte *)&buf_ptr[position + planestride * comp_num])[0] = comp>>8;
-                    ((byte *)&buf_ptr[position + planestride * comp_num])[1] = comp;
+                    ((byte *)&buf_ptr[position + planestride * (size_t)comp_num])[0] = comp>>8;
+                    ((byte *)&buf_ptr[position + planestride * (size_t)comp_num])[1] = comp;
                 }
             }
             position+=1;
@@ -907,26 +913,26 @@ gx_blend_image_buffer8to16(const byte *buf_ptr_in, unsigned short *buf_ptr_out,
         position = y * rowstride;
         for (x = 0; x < width; x++) {
             /* composite RGBA (or CMYKA, etc.) pixel with over solid background */
-            a = buf_ptr_in[position + planestride * num_comp];
+            a = buf_ptr_in[position + planestride * (size_t)num_comp];
             if (a == 0xff) {
                 for (comp_num = 0; comp_num < num_comp; comp_num++) {
-                    comp = buf_ptr_in[position + planestride * comp_num];
-                    buf_ptr_out[position + planestride * comp_num] = (comp + (comp << 8));
+                    comp = buf_ptr_in[position + planestride * (size_t)comp_num];
+                    buf_ptr_out[position + planestride * (size_t)comp_num] = (comp + (comp << 8));
                 }
             } else if (a == 0) {
                 for (comp_num = 0; comp_num < num_comp; comp_num++) {
-                    buf_ptr_out[position + planestride * comp_num] = bg_out;
+                    buf_ptr_out[position + planestride * (size_t)comp_num] = bg_out;
                 }
             } else {
                 a ^= 0xff;
                 a += (a << 8);
                 for (comp_num = 0; comp_num < num_comp; comp_num++) {
-                    comp = buf_ptr_in[position + planestride * comp_num];
+                    comp = buf_ptr_in[position + planestride * (size_t)comp_num];
                     comp += (comp << 8);
                     tmp = ((bg_out - comp) * a) + 0x8000;
                     comp += (tmp + (tmp >> 16)) >> 16;
                     comp = ((comp & 0xff) << 8) + ((comp & 0xff00) >> 8);
-                    buf_ptr_out[position + planestride * comp_num] = comp;
+                    buf_ptr_out[position + planestride * (size_t)comp_num] = comp;
                 }
             }
             position += 1;
@@ -934,262 +940,6 @@ gx_blend_image_buffer8to16(const byte *buf_ptr_in, unsigned short *buf_ptr_out,
     }
 }
 
-int
-gx_put_blended_image_cmykspot(gx_device *target, byte *buf_ptr, int planestride_in,
-                      int rowstride_in, int x0, int y0, int width, int height,
-                      int num_comp, uint16_t bg, bool has_tags, gs_int_rect rect,
-                      gs_separations * pseparations, bool deep)
-{
-    int code = 0;
-    int x, y, tmp, comp_num, output_comp_num;
-    gx_color_index color;
-    gx_color_value cv[GX_DEVICE_COLOR_MAX_COMPONENTS];
-    gx_color_value comp;
-    int input_map[GX_DEVICE_COLOR_MAX_COMPONENTS];
-    int output_map[GX_DEVICE_COLOR_MAX_COMPONENTS];
-    int num_known_comp = 0;
-    int output_num_comp = target->color_info.num_components;
-    int num_sep = pseparations->num_separations++;
-    int num_rows_left;
-    int i;
-    gx_drawing_color pdcolor;
-    gs_fixed_rect rect_fixed;
-    int bits_per_comp = ((target->color_info.depth - has_tags*8) /
-                         target->color_info.num_components);
-    bool expand = (!deep && bits_per_comp > 8);
-    int planestride = planestride_in;
-    int rowstride = rowstride_in;
-    byte *buf16_ptr = NULL;
-
-    /*
-     * The process color model for the PDF 1.4 compositor device is CMYK plus
-     * spot colors.  The target device may have only some of these colorants due
-     * to the SeparationOrder device parameter.  Thus we need to determine the
-     * mapping between the PDF 1.4 compositor and the target device.  Note:
-     * There should not be a spot colorant in the PDF 1.4 device that is not
-     * present in the target device.
-     */
-    /* Check if target processes CMYK colorants. */
-    for (comp_num = 0; comp_num < 4; comp_num++) {
-        const char * pcomp_name = (const char *)DeviceCMYKComponents[comp_num];
-
-        output_comp_num = dev_proc(target, get_color_comp_index)
-                (target, pcomp_name, strlen(pcomp_name), NO_COMP_NAME_TYPE);
-        if (output_comp_num >=0 &&
-                output_comp_num < GX_DEVICE_COLOR_MAX_COMPONENTS) {
-            output_map[num_known_comp] = output_comp_num;
-            input_map[num_known_comp++] = comp_num;
-        }
-    }
-    /* Check if target processes our spot colorants. */
-    for (comp_num = 0; comp_num < num_sep; comp_num++) {
-        output_comp_num = dev_proc(target, get_color_comp_index)
-               (target, (const char *)(pseparations->names[comp_num].data),
-                pseparations->names[comp_num].size,  NO_COMP_NAME_TYPE);
-        if (output_comp_num >= 0 &&
-                output_comp_num < GX_DEVICE_COLOR_MAX_COMPONENTS) {
-            output_map[num_known_comp] = output_comp_num;
-            input_map[num_known_comp++] = comp_num + 4;
-        }
-    }
-
-    {
-        /* See if the target device can handle the data in its current
-           form with the alpha component */
-        int alpha_offset = num_comp;
-        int tag_offset = has_tags ? num_comp + 1 : 0;
-        const byte *buf_ptrs[GS_CLIENT_COLOR_MAX_COMPONENTS];
-
-        for (i = 0; i < num_comp; i++)
-            buf_ptrs[i] = buf_ptr + i * planestride;
-        code = dev_proc(target, put_image) (target, target, buf_ptrs, num_comp,
-            rect.p.x, rect.p.y, width, height,
-            rowstride,
-            num_comp, tag_offset);
-        if (code == 0) {
-            /* Device could not handle the alpha data.  Go ahead and
-               preblend now. Note that if we do this, and we end up in the
-               default below, we only need to repack in chunky not blend.  Add
-               in conversion to 16 bits if the target device is planar and
-               a 16 bit device. */
-#if RAW_DUMP
-            /* Dump before and after the blend to make sure we are doing that ok */
-            dump_raw_buffer(target->memory, height, width, num_comp + 1, planestride, rowstride,
-                            "pre_final_blend", buf_ptr, deep);
-            global_index++;
-#endif
-            if (expand) {
-                buf16_ptr = gs_alloc_bytes(target->memory,
-                    planestride * num_comp * 2, "gx_put_blended_image_cmykspot");
-                gx_blend_image_buffer8to16(buf_ptr, (unsigned short*)buf16_ptr, width, height,
-                    rowstride, planestride, num_comp, bg>>8);
-                planestride = planestride_in * 2;
-                rowstride = rowstride_in * 2;
-                for (i = 0; i < num_comp; i++)
-                    buf_ptrs[i] = buf16_ptr + i * planestride;
-            } else {
-                if (deep) {
-                    gx_blend_image_buffer16(buf_ptr, width, height, rowstride,
-                        planestride, num_comp, bg);
-                } else {
-                    gx_blend_image_buffer(buf_ptr, width, height, rowstride,
-                        planestride, num_comp, bg>>8);
-                }
-#if RAW_DUMP
-                /* Dump before and after the blend to make sure we are doing that ok */
-                dump_raw_buffer_be(target->memory, height, width, num_comp, planestride, rowstride,
-                    "post_final_blend", buf_ptr, deep);
-                global_index++;
-                /* clist_band_count++; */
-#endif
-            }
-            /* Try again now */
-            alpha_offset = 0;
-            code = dev_proc(target, put_image) (target, target, buf_ptrs, num_comp,
-                rect.p.x, rect.p.y, width, height,
-                rowstride, alpha_offset, tag_offset);
-            if (code > 0) {
-                /* We processed some or all of the rows.  Continue until we are done */
-                num_rows_left = height - code;
-                while (num_rows_left > 0) {
-                    code = dev_proc(target, put_image) (target, target, buf_ptrs, num_comp,
-                        rect.p.x, rect.p.y + code, width,
-                        num_rows_left, rowstride,
-                        alpha_offset, tag_offset);
-                    if (code < 0) {
-                        if (buf16_ptr != NULL)
-                            gs_free_object(target->memory, buf16_ptr, "gx_put_blended_image_cmykspot");
-                        return code;
-                    }
-                    num_rows_left = num_rows_left - code;
-                }
-            }
-            if (buf16_ptr != NULL)
-                gs_free_object(target->memory, buf16_ptr, "gx_put_blended_image_cmykspot");
-            return 0;
-        }
-    }
-
-    if (buf16_ptr != NULL)
-        gs_free_object(target->memory, buf16_ptr, "gx_put_blended_image_cmykspot");
-    planestride = planestride_in;
-    rowstride = rowstride_in;
-
-    /* Clear all output colorants first */
-    for (comp_num = 0; comp_num < output_num_comp; comp_num++)
-        cv[comp_num] = 0;
-
-    /* Send pixel data to the target device. */
-    if (deep) {
-        /* NOTE: buf_ptr points to big endian data */
-        for (y = 0; y < height; y++) {
-            for (x = 0; x < width; x++) {
-
-                /* composite CMYKA, etc. pixel with over solid background */
-#define GET16_BE2NATIVE(v) \
-    ((((byte *)&(v))[0]<<8) | (((byte *)&(v))[1]))
-                uint16_t a = GET16_BE2NATIVE(buf_ptr[x + planestride * num_comp]);
-
-                if (a == 0) {
-                    for (comp_num = 0; comp_num < num_known_comp; comp_num++) {
-                        cv[output_map[comp_num]] = bg;
-                    }
-                } else if (a == 0xffff) {
-                    for (comp_num = 0; comp_num < num_known_comp; comp_num++) {
-                        comp = GET16_BE2NATIVE(buf_ptr[x + planestride * input_map[comp_num]]);
-                        cv[output_map[comp_num]] = comp;
-                    }
-                } else {
-                    /* a ^= 0xff; */  /* No inversion here! Bug 689895 */
-                    for (comp_num = 0; comp_num < num_known_comp; comp_num++) {
-                        comp = GET16_BE2NATIVE(buf_ptr[x + planestride * input_map[comp_num]]);
-                        tmp = ((comp - bg) * a) + 0x8000;
-                        comp += (tmp + (tmp >> 16))>>16;
-                        cv[output_map[comp_num]] = comp;
-                    }
-                }
-
-                /* If we have spot colors we need to encode and fill as a high level
-                   color if the device supports it which should always be the case
-                   if we are in this procedure */
-                if (dev_proc(target, dev_spec_op)(target, gxdso_supports_devn, NULL, 0)) {
-                    for (i = 0; i < output_num_comp; i++) {
-                        pdcolor.colors.devn.values[i] = cv[i];
-                    }
-                    pdcolor.type = gx_dc_type_devn;
-                    rect_fixed.p.x = int2fixed(x + x0);
-                    rect_fixed.p.y = int2fixed(y + y0);
-                    rect_fixed.q.x = int2fixed(x + x0 + 1);
-                    rect_fixed.q.y = int2fixed(y + y0 + 1);
-                    code = dev_proc(target, fill_rectangle_hl_color)(target, &rect_fixed,
-                        NULL, &pdcolor, NULL);
-                } else {
-                    /* encode as a color index */
-                    color = dev_proc(target, encode_color)(target, cv);
-                    code = dev_proc(target, fill_rectangle)(target, x + x0, y + y0, 1, 1, color);
-                }
-                if (code < 0)
-                    return code;
-            }
-
-            buf_ptr += rowstride;
-        }
-    } else {
-        bg >>= 8;
-        for (y = 0; y < height; y++) {
-            for (x = 0; x < width; x++) {
-
-                /* composite CMYKA, etc. pixel with over solid background */
-                byte a = buf_ptr[x + planestride * num_comp];
-
-                if ((a + 1) & 0xfe) {
-                    /* a ^= 0xff; */  /* No inversion here! Bug 689895 */
-                    for (comp_num = 0; comp_num < num_known_comp; comp_num++) {
-                        comp = buf_ptr[x + planestride * input_map[comp_num]];
-                        tmp = ((comp - bg) * a) + 0x80;
-                        comp += tmp + (tmp >> 8);
-                        cv[output_map[comp_num]] = comp;
-                    }
-                } else if (a == 0) {
-                    for (comp_num = 0; comp_num < num_known_comp; comp_num++) {
-                        cv[output_map[comp_num]] = bg;
-                    }
-                } else {
-                    for (comp_num = 0; comp_num < num_known_comp; comp_num++) {
-                        comp = buf_ptr[x + planestride * input_map[comp_num]];
-                        cv[output_map[comp_num]] = (comp << 8) + comp;
-                    }
-                }
-
-                /* If we have spot colors we need to encode and fill as a high level
-                   color if the device supports it which should always be the case
-                   if we are in this procedure */
-                if (dev_proc(target, dev_spec_op)(target, gxdso_supports_devn, NULL, 0)) {
-                    for (i = 0; i < output_num_comp; i++) {
-                        pdcolor.colors.devn.values[i] = cv[i];
-                    }
-                    pdcolor.type = gx_dc_type_devn;
-                    rect_fixed.p.x = int2fixed(x + x0);
-                    rect_fixed.p.y = int2fixed(y + y0);
-                    rect_fixed.q.x = int2fixed(x + x0 + 1);
-                    rect_fixed.q.y = int2fixed(y + y0 + 1);
-                    code = dev_proc(target, fill_rectangle_hl_color)(target, &rect_fixed,
-                        NULL, &pdcolor, NULL);
-                } else {
-                    /* encode as a color index */
-                    color = dev_proc(target, encode_color)(target, cv);
-                    code = dev_proc(target, fill_rectangle)(target, x + x0, y + y0, 1, 1, color);
-                }
-                if (code < 0)
-                    return code;
-            }
-
-            buf_ptr += rowstride;
-        }
-    }
-    return code;
-}
-
 int
 gx_put_blended_image_custom(gx_device *target, byte *buf_ptr_,
                       int planestride, int rowstride,
@@ -1210,7 +960,7 @@ gx_put_blended_image_custom(gx_device *target, byte *buf_ptr_,
 
                 /* composite CMYKA, etc. pixel with over solid background */
 #define GET16(v) (*((uint16_t *)(void *)&(v)))
-                uint16_t a = GET16(buf_ptr[x + planestride * num_comp]);
+                uint16_t a = GET16(buf_ptr[x + planestride * (size_t)num_comp]);
 
                 if (a == 0) {
                     for (comp_num = 0; comp_num < num_comp; comp_num++) {
@@ -1218,13 +968,13 @@ gx_put_blended_image_custom(gx_device *target, byte *buf_ptr_,
                     }
                 } else if (a == 0xffff) {
                     for (comp_num = 0; comp_num < num_comp; comp_num++) {
-                        comp = buf_ptr[x + planestride * comp_num];
+                        comp = buf_ptr[x + planestride * (size_t)comp_num];
                         cv[comp_num] = comp;
                     }
                 } else {
                     a ^= 0xffff;
                     for (comp_num = 0; comp_num < num_comp; comp_num++) {
-                        comp  = GET16(buf_ptr[x + planestride * comp_num]);
+                        comp  = GET16(buf_ptr[x + planestride * (size_t)comp_num]);
                         tmp = ((bg - comp) * a) + 0x8000;
                         cv[comp_num] = comp + ((tmp + (tmp>>16))>>16);
                     }
@@ -1244,12 +994,12 @@ gx_put_blended_image_custom(gx_device *target, byte *buf_ptr_,
             for (x = 0; x < width; x++) {
 
                 /* composite CMYKA, etc. pixel with over solid background */
-                byte a = buf_ptr[x + planestride * num_comp];
+                byte a = buf_ptr[x + planestride * (size_t)num_comp];
 
                 if ((a + 1) & 0xfe) {
                     a ^= 0xff;
                     for (comp_num = 0; comp_num < num_comp; comp_num++) {
-                        comp  = buf_ptr[x + planestride * comp_num];
+                        comp  = buf_ptr[x + planestride * (size_t)comp_num];
                         tmp = ((bg - comp) * a) + 0x80;
                         comp += tmp + (tmp >> 8);
                         cv[comp_num] = comp;
@@ -1260,7 +1010,7 @@ gx_put_blended_image_custom(gx_device *target, byte *buf_ptr_,
                     }
                 } else {
                     for (comp_num = 0; comp_num < num_comp; comp_num++) {
-                        comp = buf_ptr[x + planestride * comp_num];
+                        comp = buf_ptr[x + planestride * (size_t)comp_num];
                         cv[comp_num] = (comp << 8) + comp;
                     }
                 }
