@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2012 Artifex Software, Inc.
+/* Copyright (C) 2001-2024 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134, San Rafael,
-   CA  94903, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -31,11 +31,12 @@
 #include "gdevppla.h"
 #include "gsmemory.h"
 #include "gsmchunk.h"
-#include "gsmemlok.h"
 #include "gxclthrd.h"
 #include "gdevdevn.h"
 #include "gsicc_cache.h"
 #include "gsicc_manage.h"
+#include "gstrans.h"
+#include "gzht.h"		/* for gx_ht_cache_default_bits_size */
 
 /* Forward reference prototypes */
 static int clist_start_render_thread(gx_device *dev, int thread_index, int band);
@@ -62,6 +63,9 @@ setup_device_and_mem_for_thread(gs_memory_t *chunk_base_mem, gx_device *dev, boo
     gx_device *protodev;
     gs_c_param_list paramlist;
     gs_devn_params *pclist_devn_params;
+    gx_device_buf_space_t buf_space;
+    size_t min_buffer_space;
+
 
     /* Every thread will have a 'chunk allocator' to reduce the interaction
      * with the 'base' allocator which has 'mutex' (locking) protection.
@@ -97,11 +101,56 @@ setup_device_and_mem_for_thread(gs_memory_t *chunk_base_mem, gx_device *dev, boo
     ndev->color_info = dev->color_info;     /* copy before putdeviceparams */
     ndev->pad = dev->pad;
     ndev->log2_align_mod = dev->log2_align_mod;
-    ndev->is_planar = dev->is_planar;
-#if CMM_THREAD_SAFE
+    ndev->num_planar_planes = dev->num_planar_planes;
+    ndev->icc_struct = NULL;
+
+    /* If the device ICC profile (or proof) is OI_PROFILE, then that was not handled
+     * by put/get params, and we cannot share the profiles between the 'parent' output device
+     * and the devices created for each thread. Thus we also cannot share the icc_struct.
+     * In this case we need to create a new icc_struct and clone the profiles.  The clone
+     * operation also initializes some of the required data
+     * We need to do this *before* the gs_getdeviceparams/gs_putdeviceparams so gs_putdeviceparams
+     * will spot the same profile being used, and treat it as a no-op. Otherwise it will try to find
+     * a profile with the 'special' name "OI_PROFILE" and throw an error.
+     */
+#define DEV_PROFILE_IS(DEV, PROFILE, MATCH) \
+    ((DEV)->icc_struct != NULL &&\
+     (DEV)->icc_struct->PROFILE != NULL &&\
+     strcmp((DEV)->icc_struct->PROFILE->name, MATCH) == 0)
+
+    if (bg_print ||
+        !gscms_is_threadsafe() ||
+        DEV_PROFILE_IS(dev, device_profile[GS_DEFAULT_DEVICE_PROFILE], OI_PROFILE) ||
+        DEV_PROFILE_IS(dev, proof_profile, OI_PROFILE)) {
+        ndev->icc_struct = gsicc_new_device_profile_array(ndev);
+        if (!ndev->icc_struct) {
+            emprintf1(ndev->memory,
+                  "Error setting up device profile array, code=%d. Rendering threads not started.\n",
+                  code);
+            goto out_cleanup;
+        }
+        if ((code = gsicc_clone_profile(dev->icc_struct->device_profile[GS_DEFAULT_DEVICE_PROFILE],
+            &(ndev->icc_struct->device_profile[GS_DEFAULT_DEVICE_PROFILE]), ndev->memory)) < 0) {
+            emprintf1(dev->memory,
+                "Error setting up device profile, code=%d. Rendering threads not started.\n",
+                code);
+            goto out_cleanup;
+        }
+        if (dev->icc_struct->proof_profile &&
+           (code = gsicc_clone_profile(dev->icc_struct->proof_profile,
+            &ndev->icc_struct->proof_profile, ndev->memory)) < 0) {
+            emprintf1(dev->memory,
+                "Error setting up proof profile, code=%d. Rendering threads not started.\n",
+                code);
+            goto out_cleanup;
+
+        }
+    }
+    else {
+        /* safe to share the icc_struct among threads */
         ndev->icc_struct = dev->icc_struct;  /* Set before put params */
         rc_increment(ndev->icc_struct);
-#endif
+    }
     /* get the current device parameters to put into the cloned device */
     gs_c_param_list_write(&paramlist, thread_mem);
     if ((code = gs_getdeviceparams(dev, (gs_param_list *)&paramlist)) < 0) {
@@ -115,29 +164,6 @@ setup_device_and_mem_for_thread(gs_memory_t *chunk_base_mem, gx_device *dev, boo
         goto out_cleanup;
     gs_c_param_list_release(&paramlist);
 
-    /* If the device ICC profile (or proof) is OI_PROFILE, then that was not handled
-     * by put/get params.  In this case we need to clone the profiles.  The clone
-     * operation also initializes some of the required data */
-    if (dev->icc_struct != NULL && dev->icc_struct->device_profile[0] != NULL &&
-        strncmp(dev->icc_struct->device_profile[0]->name, OI_PROFILE, strlen(OI_PROFILE)) == 0) {
-        if ((code = gsicc_clone_profile(dev->icc_struct->device_profile[0],
-            &(ndev->icc_struct->device_profile[0]), ndev->memory)) < 0) {
-            emprintf1(dev->memory,
-                "Error setting up device profile, code=%d. Rendering threads not started.\n",
-                code);
-            goto out_cleanup;
-        }
-    }
-    if (dev->icc_struct != NULL && dev->icc_struct->proof_profile != NULL &&
-        strncmp(dev->icc_struct->proof_profile->name, OI_PROFILE, strlen(OI_PROFILE)) == 0) {
-        if ((code = gsicc_clone_profile(dev->icc_struct->proof_profile,
-            &ndev->icc_struct->proof_profile, ndev->memory)) < 0) {
-            emprintf1(dev->memory,
-                "Error setting up proof profile, code=%d. Rendering threads not started.\n",
-                code);
-            goto out_cleanup;
-        }
-    }
     /* In the case of a separation device, we need to make sure we get the
        devn params copied over */
     pclist_devn_params = dev_proc(dev, ret_devn_params)(dev);
@@ -154,14 +180,31 @@ setup_device_and_mem_for_thread(gs_memory_t *chunk_base_mem, gx_device *dev, boo
     ndev->icc_struct->supports_devn = cdev->icc_struct->supports_devn;
     ncdev->page_uses_transparency = cdev->page_uses_transparency;
     if_debug3m(gs_debug_flag_icc, cdev->memory,
-               "[icc] MT clist device = 0x%p profile = 0x%p handle = 0x%p\n",
-               ncdev,
-               ncdev->icc_struct->device_profile[0],
-               ncdev->icc_struct->device_profile[0]->profile_handle);
+               "[icc] MT clist device = "PRI_INTPTR" profile = "PRI_INTPTR" handle = "PRI_INTPTR"\n",
+               (intptr_t)ncdev,
+               (intptr_t)ncdev->icc_struct->device_profile[GS_DEFAULT_DEVICE_PROFILE],
+               (intptr_t)ncdev->icc_struct->device_profile[GS_DEFAULT_DEVICE_PROFILE]->profile_handle);
     /* If the device is_planar, then set the flag in the new_device and the procs */
-    if ((ncdev->is_planar = cdev->is_planar))
+    if ((ncdev->num_planar_planes = cdev->num_planar_planes))
         gdev_prn_set_procs_planar(ndev);
 
+    /* Make sure that the ncdev BandHeight matches what we used when writing the clist, but
+     * re-calculate the BandBufferSpace so we don't over-allocate (in case the page uses
+     * transparency so that the BandHeight was reduced.)
+     */
+    ncdev->space_params.band = cdev->page_info.band_params;
+    ncdev->space_params.banding_type = BandingAlways;
+    code = npdev->printer_procs.buf_procs.size_buf_device
+                (&buf_space, (gx_device *)ncdev, NULL, ncdev->space_params.band.BandHeight, false);
+    min_buffer_space = clist_minimum_buffer(cdev->nbands);
+    ncdev->space_params.band.BandBufferSpace = buf_space.bits + buf_space.line_ptrs;
+    /* Check if the BandBufferSpace is large enough to allow us for clist writing */
+    /* to prevent an error from gdev_prn_allocate_memory which checks that.       */
+    if (min_buffer_space > ncdev->space_params.band.BandBufferSpace)
+        ncdev->space_params.band.BandBufferSpace = min_buffer_space;
+    ncdev->space_params.band.tile_cache_size = cdev->page_info.tile_cache_size;	/* must be the same */
+    ncdev->space_params.band.BandBufferSpace += cdev->page_info.tile_cache_size;
+
     /* gdev_prn_allocate_memory sets the clist for writing, creating new files.
      * We need  to unlink those files and open the main thread's files, then
      * reset the clist state for reading/rendering
@@ -169,6 +212,13 @@ setup_device_and_mem_for_thread(gs_memory_t *chunk_base_mem, gx_device *dev, boo
     if ((code = gdev_prn_allocate_memory(ndev, NULL, ndev->width, ndev->height)) < 0)
         goto out_cleanup;
 
+    if (ncdev->page_info.tile_cache_size != cdev->page_info.tile_cache_size) {
+        emprintf2(thread_mem,
+                   "clist_setup_render_threads: tile_cache_size mismatch. New size=%"PRIdSIZE", should be %"PRIdSIZE"\n",
+                   ncdev->page_info.tile_cache_size, cdev->page_info.tile_cache_size);
+        goto out_cleanup;
+    }
+
     /* close and unlink the temp files just created */
     ncdev->page_info.io_procs->fclose(ncdev->page_info.cfile, ncdev->page_info.cfname, true);
     ncdev->page_info.io_procs->fclose(ncdev->page_info.bfile, ncdev->page_info.bfname, true);
@@ -191,23 +241,23 @@ setup_device_and_mem_for_thread(gs_memory_t *chunk_base_mem, gx_device *dev, boo
     /* The threads are maintained until clist_finish_page.  At which
        point, the threads are torn down, the master clist reader device
        is changed to writer, and the icc_table and the icc_cache_cl freed */
-#if CMM_THREAD_SAFE
+    if (dev->icc_struct == ndev->icc_struct) {
     /* safe to share the link cache */
-    ncdev->icc_cache_cl = cdev->icc_cache_cl;
-    rc_increment(cdev->icc_cache_cl, "setup_render_thread");
-#else
-    /* each thread needs its own link cache */
-    if (cachep != NULL) {
-        if (*cachep == NULL) {
-            /* We don't have one cached that we can reuse, so make one. */
-            if ((*cachep = gsicc_cache_new(thread_mem->thread_safe_memory)) == NULL)
-                goto out_cleanup;
-        }
-        rc_increment(*cachep);
-        ncdev->icc_cache_cl = *cachep;
-    } else if ((ncdev->icc_cache_cl = gsicc_cache_new(thread_mem)) == NULL)
-        goto out_cleanup;
-#endif
+        ncdev->icc_cache_cl = cdev->icc_cache_cl;
+        rc_increment(cdev->icc_cache_cl);		/* FIXME: needs to be incdemented safely */
+    } else {
+        /* each thread needs its own link cache */
+        if (cachep != NULL) {
+            if (*cachep == NULL) {
+                /* We don't have one cached that we can reuse, so make one. */
+                if ((*cachep = gsicc_cache_new(thread_mem->thread_safe_memory)) == NULL)
+                    goto out_cleanup;
+            }
+            rc_increment(*cachep);
+                ncdev->icc_cache_cl = *cachep;
+        } else if ((ncdev->icc_cache_cl = gsicc_cache_new(thread_mem->thread_safe_memory)) == NULL)
+            goto out_cleanup;
+    }
     if (bg_print) {
         gx_device_clist_reader *ncrdev = (gx_device_clist_reader *)ncdev;
 
@@ -264,30 +314,65 @@ clist_setup_render_threads(gx_device *dev, int y, gx_process_page_options_t *opt
     gs_memory_t *mem = cdev->bandlist_memory;
     gs_memory_t *chunk_base_mem = mem->thread_safe_memory;
     gs_memory_status_t mem_status;
-    int i, band;
+    int i, j, band;
     int code = 0;
     int band_count = cdev->nbands;
     int band_height = crdev->page_info.band_params.BandHeight;
+    byte **reserve_memory_array = NULL;
+    int reserve_pdf14_memory_size = 0;
+    /* space for the halftone cache plus 2Mb for other allocations during rendering (paths, etc.) */
+    /* this will be increased by the measured profile storage and icclinks (estimated).		  */
+    int reserve_size = 2 * 1024 * 1024 + (gx_ht_cache_default_bits_size() * dev->color_info.num_components);
+    clist_icctable_entry_t *curr_entry;
+    bool deep = device_is_deep(dev);
+
     crdev->num_render_threads = pdev->num_render_threads_requested;
 
     if(gs_debug[':'] != 0)
         dmprintf1(mem, "%% %d rendering threads requested.\n", pdev->num_render_threads_requested);
 
+    if (crdev->page_uses_transparency) {
+        reserve_pdf14_memory_size = (ESTIMATED_PDF14_ROW_SPACE(max(1, crdev->width), crdev->color_info.num_components, deep ? 16 : 8) >> 3);
+        reserve_pdf14_memory_size *= crdev->page_info.band_params.BandHeight;	/* BandHeight set by writer */
+    }
+    /* scan the profile table sizes to get the total each thread will need */
+    if (crdev->icc_table != NULL) {
+        for (curr_entry = crdev->icc_table->head; curr_entry != NULL; curr_entry = curr_entry->next) {
+            reserve_size += curr_entry->serial_data.size;
+            /* FIXME: Should actually measure the icclink size to device (or pdf14 blend space) */
+            reserve_size += 2 * 1024 * 1024;		/* a worst case estimate */
+        }
+    }
     if (crdev->num_render_threads > band_count)
         crdev->num_render_threads = band_count; /* don't bother starting more threads than bands */
+    /* don't exceed our limit (allow for BGPrint and main thread) */
+    if (crdev->num_render_threads > MAX_THREADS - 2)
+        crdev->num_render_threads = MAX_THREADS - 2;
 
     /* Allocate and initialize an array of thread control structures */
     crdev->render_threads = (clist_render_thread_control_t *)
               gs_alloc_byte_array(mem, crdev->num_render_threads,
-              sizeof(clist_render_thread_control_t), "clist_setup_render_threads" );
+                                  sizeof(clist_render_thread_control_t),
+                                  "clist_setup_render_threads");
     /* fallback to non-threaded if allocation fails */
     if (crdev->render_threads == NULL) {
         emprintf(mem, " VMerror prevented threads from starting.\n");
         return_error(gs_error_VMerror);
     }
-
+    reserve_memory_array = (byte **)gs_alloc_byte_array(mem,
+                                                        crdev->num_render_threads,
+                                                        sizeof(void *),
+                                                        "clist_setup_render_threads");
+    if (reserve_memory_array == NULL) {
+        gs_free_object(mem, crdev->render_threads, "clist_setup_render_threads");
+        crdev->render_threads = NULL;
+        emprintf(mem, " VMerror prevented threads from starting.\n");
+        return_error(gs_error_VMerror);
+    }
+    memset(reserve_memory_array, 0, crdev->num_render_threads * sizeof(void *));
     memset(crdev->render_threads, 0, crdev->num_render_threads *
             sizeof(clist_render_thread_control_t));
+
     crdev->main_thread_data = cdev->data;               /* save data area */
     /* Based on the line number requested, decide the order of band rendering */
     /* Almost all devices go in increasing line order (except the bmp* devices ) */
@@ -302,7 +387,7 @@ clist_setup_render_threads(gx_device *dev, int y, gx_process_page_options_t *opt
 
     /* If we don't have one large enough already, create an icc cache list */
     if (crdev->num_render_threads > crdev->icc_cache_list_len) {
-        void *old = &(crdev->icc_cache_list[0]);
+        gsicc_link_cache_t **old = crdev->icc_cache_list;
         crdev->icc_cache_list = (gsicc_link_cache_t **)gs_alloc_byte_array(mem->thread_safe_memory,
                                     crdev->num_render_threads,
                                     sizeof(void*), "clist_render_setup_threads");
@@ -311,7 +396,7 @@ clist_setup_render_threads(gx_device *dev, int y, gx_process_page_options_t *opt
             return_error(gs_error_VMerror);
         }
         if (crdev->icc_cache_list_len > 0)
-            memcpy(crdev->icc_cache_list, old, crdev->icc_cache_list_len * sizeof(void *));
+            memcpy(crdev->icc_cache_list, old, crdev->icc_cache_list_len * sizeof(gsicc_link_cache_t *));
         memset(&(crdev->icc_cache_list[crdev->icc_cache_list_len]), 0,
             (crdev->num_render_threads - crdev->icc_cache_list_len) * sizeof(void *));
         crdev->icc_cache_list_len = crdev->num_render_threads;
@@ -324,6 +409,14 @@ clist_setup_render_threads(gx_device *dev, int y, gx_process_page_options_t *opt
         gx_device *ndev;
         clist_render_thread_control_t *thread = &(crdev->render_threads[i]);
 
+        /* arbitrary extra reserve for other allocation by threads (paths, etc.) */
+        /* plus the amount estimated for the pdf14 buffers */
+        reserve_memory_array[i] = (byte *)gs_alloc_bytes(mem, reserve_size + reserve_pdf14_memory_size,
+                                                         "clist_render_setup_threads");
+        if (reserve_memory_array[i] == NULL) {
+            code = gs_error_VMerror;	/* set code to an error for cleanup after the loop */
+        break;
+        }
         ndev = setup_device_and_mem_for_thread(chunk_base_mem, dev, false, &crdev->icc_cache_list[i]);
         if (ndev == NULL) {
             code = gs_error_VMerror;	/* set code to an error for cleanup after the loop */
@@ -344,21 +437,21 @@ clist_setup_render_threads(gx_device *dev, int y, gx_process_page_options_t *opt
         /* create the buf device for this thread, and allocate the semaphores */
         if ((code = gdev_create_buf_device(cdev->buf_procs.create_buf_device,
                                 &(thread->bdev), ndev,
-                                band*crdev->page_band_height, NULL,
-                                thread->memory, &(crdev->color_usage_array[0])) < 0))
+                                band*crdev->page_info.band_params.BandHeight, NULL,
+                                thread->memory, &(crdev->color_usage_array[0]))) < 0)
             break;
         if ((thread->sema_this = gx_semaphore_label(gx_semaphore_alloc(thread->memory), "Band")) == NULL ||
             (thread->sema_group = gx_semaphore_label(gx_semaphore_alloc(thread->memory), "Group")) == NULL) {
             code = gs_error_VMerror;
             break;
         }
-        /* Start thread 'i' to do band */
-        if ((code = clist_start_render_thread(dev, i, band)) < 0)
-            break;
+        /* We don't start the threads yet until we  free up the */
+        /* reserve memory we have allocated for that band. */
+        thread->band = band;
     }
     /* If the code < 0, the last thread creation failed -- clean it up */
     if (code < 0) {
-        band -= crdev->thread_lookahead_direction;	/* update for 'next_band' usage */
+        /* NB: 'band' will be the one that failed, so will be the next_band needed to start */
         /* the following relies on 'free' ignoring NULL pointers */
         gx_semaphore_free(crdev->render_threads[i].sema_group);
         gx_semaphore_free(crdev->render_threads[i].sema_this);
@@ -391,9 +484,7 @@ clist_setup_render_threads(gx_device *dev, int y, gx_process_page_options_t *opt
     if (i == 0) {
         if (crdev->render_threads[0].memory != NULL) {
             gs_memory_chunk_release(crdev->render_threads[0].memory);
-            /* free up the locking wrapper if we allocated one */
             if (chunk_base_mem != mem) {
-                gs_memory_locked_release((gs_memory_locked_t *)chunk_base_mem);
                 gs_free_object(mem, chunk_base_mem, "clist_setup_render_threads(locked allocator)");
             }
         }
@@ -415,14 +506,24 @@ clist_setup_render_threads(gx_device *dev, int y, gx_process_page_options_t *opt
         emprintf1(mem, "Rendering threads not started, code=%d.\n", code);
         return_error(code);
     }
+    /* Free up any "reserve" memory we may have allocated, and start the
+     * threads since we deferred that in the thread setup loop above.
+     * We know if we get here we can start at least 1 thread.
+     */
+    for (j=0, code = 0; j<crdev->num_render_threads; j++) {
+        gs_free_object(mem, reserve_memory_array[j], "clist_setup_render_threads");
+        if (code == 0 && j < i)
+            code = clist_start_render_thread(dev, j, crdev->render_threads[j].band);
+    }
+    gs_free_object(mem, reserve_memory_array, "clist_setup_render_threads");
     crdev->num_render_threads = i;
     crdev->curr_render_thread = 0;
-        crdev->next_band = band;
+    crdev->next_band = band;
 
     if(gs_debug[':'] != 0)
         dmprintf1(mem, "%% Using %d rendering threads\n", i);
 
-    return 0;
+    return code;
 }
 
 /* This is also exported for teardown after background printing */
@@ -498,13 +599,18 @@ clist_teardown_render_threads(gx_device *dev)
     int i;
 
     if (crdev->render_threads != NULL) {
-        /* Wait for each thread to finish then free its memory */
+        /* Wait for all threads to finish */
+        for (i = (crdev->num_render_threads - 1); i >= 0; i--) {
+            clist_render_thread_control_t *thread = &(crdev->render_threads[i]);
+
+            if (thread->status == THREAD_BUSY)
+                gx_semaphore_wait(thread->sema_this);
+        }
+        /* then free each thread's memory */
         for (i = (crdev->num_render_threads - 1); i >= 0; i--) {
             clist_render_thread_control_t *thread = &(crdev->render_threads[i]);
             gx_device_clist_common *thread_cdev = (gx_device_clist_common *)thread->cdev;
 
-            if (thread->status == THREAD_BUSY)
-                gx_semaphore_wait(thread->sema_this);
             /* Free control semaphores */
             gx_semaphore_free(thread->sema_group);
             gx_semaphore_free(thread->sema_this);
@@ -578,11 +684,11 @@ clist_render_thread(void *data)
     gx_device_clist_reader *crdev = &cldev->reader;
     gx_device *bdev = thread->bdev;
     gs_int_rect band_rect;
-    byte *mdata = crdev->data + crdev->page_tile_cache_size;
-    byte *mlines = (crdev->page_line_ptrs_offset == 0 ? NULL : mdata + crdev->page_line_ptrs_offset);
+    byte *mdata = crdev->data + crdev->page_info.tile_cache_size;
+    byte *mlines = (crdev->page_info.line_ptrs_offset == 0 ? NULL : mdata + crdev->page_info.line_ptrs_offset);
     uint raster = gx_device_raster_plane(dev, NULL);
     int code;
-    int band_height = crdev->page_band_height;
+    int band_height = crdev->page_info.band_params.BandHeight;
     int band = thread->band;
     int band_begin_line = band * band_height;
     int band_end_line = band_begin_line + band_height;
@@ -657,6 +763,10 @@ clist_get_band_from_thread(gx_device *dev, int band_needed, gx_process_page_opti
     if (thread->band != band_needed) {
         int band = band_needed;
 
+        emprintf3(thread->memory,
+                  "thread->band = %d, band_needed = %d, direction = %d, ",
+                  thread->band, band_needed, crdev->thread_lookahead_direction);
+
         /* Probably we went in the wrong direction, so let the threads */
         /* all complete, then restart them in the opposite direction   */
         /* If the caller is 'bouncing around' we may end up back here, */
@@ -672,7 +782,10 @@ clist_get_band_from_thread(gx_device *dev, int band_needed, gx_process_page_opti
             crdev->thread_lookahead_direction = -1;   /* assume backwards if we are asking for the last band */
         if (band_needed == 0)
             crdev->thread_lookahead_direction = 1;    /* force forward if we are looking for band 0 */
-        /* Loop creating the devices and semaphores for each thread, then start them */
+
+        dmprintf1(thread->memory, "new_direction = %d\n", crdev->thread_lookahead_direction);
+
+        /* Loop starting the threads in the new lookahead_direction */
         for (i=0; (i < crdev->num_render_threads) && (band >= 0) && (band < band_count);
                 i++, band += crdev->thread_lookahead_direction) {
             thread = &(crdev->render_threads[i]);
@@ -727,7 +840,7 @@ clist_get_band_from_thread(gx_device *dev, int band_needed, gx_process_page_opti
 /* rendering adjacent to the first band (forward or backward) */
 static int
 clist_get_bits_rect_mt(gx_device *dev, const gs_int_rect * prect,
-                         gs_get_bits_params_t *params, gs_int_rect **unread)
+                         gs_get_bits_params_t *params)
 {
     gx_device_printer *pdev = (gx_device_printer *)dev;
     gx_device_clist *cldev = (gx_device_clist *)dev;
@@ -751,7 +864,7 @@ clist_get_bits_rect_mt(gx_device *dev, const gs_int_rect * prect,
     /* This page might not want multiple threads */
     /* Also we don't support plane extraction using multiple threads */
     if (pdev->num_render_threads_requested < 1 || (options & GB_SELECT_PLANES))
-        return clist_get_bits_rectangle(dev, prect, params, unread);
+        return clist_get_bits_rectangle(dev, prect, params);
 
     if (prect->p.x < 0 || prect->q.x > dev->width ||
         y < 0 || end_y > dev->height
@@ -768,12 +881,12 @@ clist_get_bits_rect_mt(gx_device *dev, const gs_int_rect * prect,
         /* Haven't done any rendering yet, try to set up the threads */
         if (clist_setup_render_threads(dev, y, NULL) < 0)
             /* problem setting up the threads, revert to single threaded */
-            return clist_get_bits_rectangle(dev, prect, params, unread);
+            return clist_get_bits_rectangle(dev, prect, params);
     } else {
         if (crdev->render_threads == NULL) {
             /* If we get here with with ymin and ymax > 0 it's because we closed the threads */
             /* while doing a page due to an error. Use single threaded mode.         */
-            return clist_get_bits_rectangle(dev, prect, params, unread);
+            return clist_get_bits_rectangle(dev, prect, params);
         }
     }
     /* If we already have the band's data, just return it */
@@ -781,7 +894,7 @@ clist_get_bits_rect_mt(gx_device *dev, const gs_int_rect * prect,
         code = clist_get_band_from_thread(dev, band, NULL);
     if (code < 0)
         goto free_thread_out;
-    mdata = crdev->data + crdev->page_tile_cache_size;
+    mdata = crdev->data + crdev->page_info.tile_cache_size;
     if ((code = gdev_create_buf_device(cdev->buf_procs.create_buf_device,
                                   &bdev, cdev->target, y, NULL,
                                   mem, &(crdev->color_usage_array[band]))) < 0 ||
@@ -795,7 +908,7 @@ clist_get_bits_rect_mt(gx_device *dev, const gs_int_rect * prect,
     band_rect.p.y = 0;
     band_rect.q.y = lines_rasterized;
     code = dev_proc(bdev, get_bits_rectangle)
-        (bdev, &band_rect, params, unread);
+        (bdev, &band_rect, params);
     cdev->buf_procs.destroy_buf_device(bdev);
     if (code < 0)
         goto free_thread_out;
@@ -814,7 +927,7 @@ clist_get_bits_rect_mt(gx_device *dev, const gs_int_rect * prect,
      * rectangles, punt.
      */
     if (!(options & GB_RETURN_COPY) || code > 0)
-        return gx_default_get_bits_rectangle(dev, prect, params, unread);
+        return gx_default_get_bits_rectangle(dev, prect, params);
     options = params->options;
     if (!(options & GB_RETURN_COPY)) {
         /* Redo the first piece with copying. */
@@ -842,7 +955,7 @@ clist_get_bits_rect_mt(gx_device *dev, const gs_int_rect * prect,
             band_rect.p.y = my;
             band_rect.q.y = my + lines_rasterized;
             code = dev_proc(bdev, get_bits_rectangle)
-                (bdev, &band_rect, &band_params, unread);
+                (bdev, &band_rect, &band_params);
             if (code < 0)
                 break;
             params->options = band_params.options;
@@ -867,7 +980,7 @@ clist_process_page(gx_device *dev, gx_process_page_options_t *options)
     gx_device_clist_common *cdev = (gx_device_clist_common *)dev;
     int y;
     int line_count;
-    int band_height = crdev->page_band_height;
+    int band_height = crdev->page_info.band_params.BandHeight;
     gs_int_rect band_rect;
     int lines_rasterized;
     gx_device *bdev;
@@ -885,7 +998,7 @@ clist_process_page(gx_device *dev, gx_process_page_options_t *options)
             return code;
     }
 
-    render_plane.index = -1;
+    gx_render_plane_init(&render_plane, dev, -1);
     for (y = 0; y < dev->height; y += lines_rasterized)
     {
         line_count = band_height;
