@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2018 Artifex Software, Inc.
+/* Copyright (C) 2001-2024 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,13 +9,14 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
 /* Implementation of clipping paths, other than actual clipping */
 #include "gx.h"
+#include "string_.h"
 #include "gserrors.h"
 #include "gsstruct.h"
 #include "gsutil.h"
@@ -69,18 +70,20 @@ RELOC_PTRS_END
 static
 ENUM_PTRS_WITH(device_clip_enum_ptrs, gx_device_clip *cptr)
 {
-    if (index < st_clip_list_max_ptrs + 2)
+    if (index < st_clip_list_max_ptrs + 3)
         return ENUM_USING(st_clip_list, &cptr->list,
-                          sizeof(gx_clip_list), index - 2);
+                          sizeof(gx_clip_list), index - 3);
     return ENUM_USING(st_device_forward, vptr,
                       sizeof(gx_device_forward),
-                      index - (st_clip_list_max_ptrs + 2));
+                      index - (st_clip_list_max_ptrs + 3));
 }
 case 0:
 ENUM_RETURN((cptr->current == &cptr->list.single ? NULL :
              (void *)cptr->current));
 case 1:
 ENUM_RETURN((cptr->cpath));
+case 2:
+ENUM_RETURN((cptr->rect_list));
 ENUM_PTRS_END
 static
 RELOC_PTRS_WITH(device_clip_reloc_ptrs, gx_device_clip *cptr)
@@ -90,6 +93,7 @@ RELOC_PTRS_WITH(device_clip_reloc_ptrs, gx_device_clip *cptr)
     else
         RELOC_PTR(gx_device_clip, current);
     RELOC_PTR(gx_device_clip, cpath);
+    RELOC_PTR(gx_device_clip, rect_list);
     RELOC_USING(st_clip_list, &cptr->list, sizeof(gx_clip_list));
     RELOC_USING(st_device_forward, vptr, sizeof(gx_device_forward));
 }
@@ -131,6 +135,8 @@ cpath_init_rectangle(gx_clip_path * pcpath, gs_fixed_rect * pbox)
     gx_clip_list_from_rectangle(&pcpath->rect_list->list, pbox);
     pcpath->inner_box = *pbox;
     pcpath->path_valid = false;
+    pcpath->path_fill_adjust.x = 0;
+    pcpath->path_fill_adjust.y = 0;
     pcpath->path.bbox = *pbox;
     gx_cpath_set_outer_box(pcpath);
     pcpath->id = gs_next_ids(pcpath->path.memory, 1);	/* path changed => change id */
@@ -150,6 +156,7 @@ cpath_share_own_contents(gx_clip_path * pcpath, const gx_clip_path * shared)
 {
     pcpath->inner_box = shared->inner_box;
     pcpath->path_valid = shared->path_valid;
+    pcpath->path_fill_adjust = shared->path_fill_adjust;
     pcpath->outer_box = shared->outer_box;
     pcpath->id = shared->id;
     pcpath->cached = NULL;
@@ -171,8 +178,10 @@ gx_cpath_init_contained_shared(gx_clip_path * pcpath,
 {
     if (shared) {
         if (shared->path.segments == &shared->path.local_segments) {
-            lprintf1("Attempt to share (local) segments of clip path 0x%lx!\n",
-                     (ulong) shared);
+#ifdef DEBUG
+            lprintf1("Attempt to share (local) segments of clip path "PRI_INTPTR"!\n",
+                     (intptr_t)shared);
+#endif
             return_error(gs_error_Fatal);
         }
         *pcpath = *shared;
@@ -229,8 +238,10 @@ gx_cpath_init_local_shared_nested(gx_clip_path * pcpath,
     if (shared) {
         if ((shared->path.segments == &shared->path.local_segments) &&
             !safely_nested) {
-            lprintf1("Attempt to share (local) segments of clip path 0x%lx!\n",
-                     (ulong) shared);
+#ifdef DEBUG
+            lprintf1("Attempt to share (local) segments of clip path "PRI_INTPTR"!\n",
+                     (intptr_t)shared);
+#endif
             return_error(gs_error_Fatal);
         }
         pcpath->path = shared->path;
@@ -275,6 +286,8 @@ gx_cpath_unshare(gx_clip_path * pcpath)
             return code;
         /* Copy the rectangle list. */
 /**************** NYI ****************/
+        /* Until/Unless we implement this, NULL out the list */
+        memset(&pcpath->rect_list->list, 0x00, sizeof(gx_clip_list));
         rc_decrement(rlist, "gx_cpath_unshare");
     }
     return code;
@@ -322,8 +335,10 @@ gx_cpath_assign_preserve(gx_clip_path * pcpto, gx_clip_path * pcpfrom)
             int code = cpath_alloc_list(&tolist, tolist->rc.memory,
                                         "gx_cpath_assign");
 
-            if (code < 0)
+            if (code < 0) {
+                rc_decrement(pcpto->path.segments, "gx_path_assign");
                 return code;
+            }
             rc_decrement(pcpto->rect_list, "gx_cpath_assign");
         } else {
             /* Use pcpto's list object. */
@@ -339,6 +354,7 @@ gx_cpath_assign_preserve(gx_clip_path * pcpto, gx_clip_path * pcpfrom)
         rc_decrement(pcpto->rect_list, "gx_cpath_assign");
     }
     rc_increment(pcpfrom->path_list);
+    rc_decrement(pcpto->path_list, "gx_cpath_assign");
     path = pcpto->path, *pcpto = *pcpfrom, pcpto->path = path;
     return 0;
 }
@@ -350,7 +366,7 @@ gx_cpath_assign_free(gx_clip_path * pcpto, gx_clip_path * pcpfrom)
     int code = gx_cpath_assign_preserve(pcpto, pcpfrom);
 
     if (code < 0)
-        return 0;
+        return code;
     gx_cpath_free(pcpfrom, "gx_cpath_assign_free");
     return 0;
 }
@@ -397,8 +413,10 @@ gx_cpath_path_list_new(gs_memory_t *mem, gx_clip_path *pcpath, int rule,
     rc_init_free(pcplist, mem, 1, rc_free_cpath_path_list);
     if (pcpath!=NULL && !pcpath->path_valid) {
         code = gx_path_init_contained_shared(&pcplist->path, NULL, mem, cname);
-        if (code < 0)
+        if (code < 0) {
+            gs_free_object(mem, pcplist, "gx_cpath_path_list_new");
             return code;
+        }
         code = gx_cpath_to_path(pcpath, &pcplist->path);
     } else {
         gx_path_init_local(&pcplist->path, mem);
@@ -407,6 +425,7 @@ gx_cpath_path_list_new(gs_memory_t *mem, gx_clip_path *pcpath, int rule,
     if (code < 0)
         return code;
     pcplist->next = next;
+    rc_increment(next);
     pcplist->rule = rule;
     *pnew = pcplist;
     return 0;
@@ -475,6 +494,8 @@ gx_cpath_to_path(gx_clip_path * pcpath, gx_path * ppath)
         if (code < 0)
             return code;
         pcpath->path_valid = true;
+        pcpath->path_fill_adjust.x = 0;
+        pcpath->path_fill_adjust.y = 0;
     }
     return gx_path_assign_preserve(ppath, &pcpath->path);
 }
@@ -548,9 +569,11 @@ cpath_set_rectangle(gx_clip_path * pcpath, gs_fixed_rect * pbox)
         int code = cpath_alloc_list(&pcpath->rect_list, pcpath->path.memory,
                                     "gx_cpath_from_rectangle");
 
-        rc_decrement(rlist, "gx_cpath_from_rectangle");
-        if (code < 0)
+        if (code < 0) {
+            pcpath->rect_list = rlist;
             return code;
+        }
+        rc_decrement(rlist, "gx_cpath_from_rectangle");
         rlist = pcpath->rect_list;
     }
     cpath_init_rectangle(pcpath, pbox);
@@ -696,6 +719,7 @@ gx_cpath_intersect_with_params(gx_clip_path *pcpath, /*const*/ gx_path *ppath_or
             /* The path is valid; otherwise, defer constructing it. */
             gx_path_assign_preserve(&pcpath->path, ppath);
             pcpath->path_valid = true;
+            pcpath->path_fill_adjust = params != NULL ? params->adjust : pgs->fill_adjust;
         }
     } else {
         /* New clip path is nontrivial.  Intersect the slow way. */
@@ -717,19 +741,24 @@ gx_cpath_intersect_with_params(gx_clip_path *pcpath, /*const*/ gx_path *ppath_or
             /* gx_cpath_intersect_path_slow NULLs pcpath->path_list, so
              * remember it here. */
             next = pcpath->path_list;
+            rc_increment(next);
         }
         code = gx_cpath_intersect_path_slow(pcpath, (params != NULL ? ppath_orig : ppath),
                             rule, pgs, params);
-        if (code < 0)
+        if (code < 0) {
+            rc_decrement(next, "gx_cpath_clip");
             goto ex;
+        }
         if (path_valid) {
             gx_path_assign_preserve(&pcpath->path, ppath_orig);
             pcpath->path_valid = true;
+            pcpath->path_fill_adjust = params != NULL ? params->adjust : pgs->fill_adjust;
             pcpath->rule = rule;
         } else {
             code = gx_cpath_path_list_new(pcpath->path.memory, NULL, rule,
                                           ppath_orig, next, &pcpath->path_list);
         }
+        rc_decrement(next, "gx_cpath_clip");
     }
 ex:
     if (ppath != ppath_orig)
@@ -1109,7 +1138,11 @@ gx_cpath_copy(const gx_clip_path * from, gx_clip_path * pcpath)
     pcpath->cached = NULL;
     l->single = from->rect_list->list.single;
     for (r = from->rect_list->list.head; r != NULL; r = r->next) {
-        s = gs_alloc_struct(from->rect_list->rc.memory, gx_clip_rect, &st_clip_rect, "gx_cpath_copy");
+        if (pcpath->rect_list->rc.memory == NULL)
+            s = gs_alloc_struct(from->rect_list->rc.memory, gx_clip_rect, &st_clip_rect, "gx_cpath_copy");
+        else
+            s = gs_alloc_struct(pcpath->rect_list->rc.memory, gx_clip_rect, &st_clip_rect, "gx_cpath_copy");
+
         if (s == NULL)
             return_error(gs_error_VMerror);
         *s = *r;
