@@ -1,14 +1,16 @@
-/* Copyright (C) 2001-2018 Artifex Software, Inc.
+/* Copyright (C) 2001-2024 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
    implied.
 
-   This software is distributed under license and may not be copied, modified
-   or distributed except as expressly authorized under the terms of that
-   license.  Refer to licensing information at http://www.artifex.com/
-   or contact Artifex Software, Inc.,  1305 Grant Avenue - Suite 200,
-   Novato, CA 94945, U.S.A., +1(415)492-9861, for further information.
+   This software is distributed under license and may not be copied,
+   modified or distributed except as expressly authorized under the terms
+   of the license contained in the file LICENSE in this distribution.
+
+   Refer to licensing information at http://www.artifex.com or contact
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -33,9 +35,13 @@
 #include "gsimage.h"
 #include "gsbittab.h"
 #include "gzpath.h"
+#include "gxdevsop.h"
 
 #include "gxfapi.h"
 
+#define FAPI_ROUND(v) (v >= 0 ? v + 0.5 : v - 0.5)
+#define FAPI_ROUND_TO_FRACINT(v) ((fracint)FAPI_ROUND(v))
+
 extern_gs_get_fapi_server_inits();
 
 /* FIXME */
@@ -55,6 +61,7 @@ gs_fapi_renderer_retcode(gs_memory_t *mem, gs_fapi_server *I,
 
 typedef struct gs_fapi_outline_handler_s
 {
+    gs_fapi_server *fserver;
     struct gx_path_s *path;
     fixed x0;
     fixed y0;
@@ -97,29 +104,33 @@ add_move(gs_fapi_path *I, int64_t x, int64_t y)
 {
     gs_fapi_outline_handler *olh = (gs_fapi_outline_handler *) I->olh;
 
-    x = import_shift(x, I->shift) + olh->x0;
-    y = -import_shift(y, I->shift) + olh->y0;
-
-    if (x > (int64_t) max_fixed) {
-        x = (int64_t) max_fixed;
-    }
-    else if (x < (int64_t) min_fixed) {
-        x = (int64_t) min_fixed;
+    x = import_shift(x, I->shift);
+    y = -import_shift(y, I->shift);
+    if (olh->fserver->transform_outline) {
+        gs_point pt;
+        I->gs_error = gs_distance_transform((double)fixed2float((float)x), (double)fixed2float((float)y), &olh->fserver->outline_mat, &pt);
+        if (I->gs_error < 0)
+            return I->gs_error;
+        x = float2fixed(pt.x);
+        y = float2fixed(pt.y);
     }
+    x += olh->x0;
+    y += olh->y0;
 
-    if (y > (int64_t) max_fixed) {
-        y = (int64_t) max_fixed;
+    if (x > (int64_t) max_coord_fixed || x < (int64_t) min_coord_fixed
+     || y > (int64_t) max_coord_fixed || y < (int64_t) min_coord_fixed) {
+         I->gs_error = gs_error_undefinedresult;
     }
-    else if (y < (int64_t) min_fixed) {
-        y = (int64_t) min_fixed;
+    else {
+
+        if (olh->need_close && olh->close_path)
+            if ((I->gs_error = add_closepath(I)) < 0)
+                return (I->gs_error);
+        olh->need_close = false;
+
+/*        dprintf2("%f %f moveto\n", fixed2float(x), fixed2float(y)); */
+        I->gs_error = gx_path_add_point(olh->path, (fixed) x, (fixed) y);
     }
-
-    if (olh->need_close && olh->close_path)
-        if ((I->gs_error = add_closepath(I)) < 0)
-            return (I->gs_error);
-    olh->need_close = false;
-    I->gs_error = gx_path_add_point(olh->path, (fixed) x, (fixed) y);
-
     return (I->gs_error);
 }
 
@@ -128,24 +139,29 @@ add_line(gs_fapi_path *I, int64_t x, int64_t y)
 {
     gs_fapi_outline_handler *olh = (gs_fapi_outline_handler *) I->olh;
 
-    x = import_shift(x, I->shift) + olh->x0;
-    y = -import_shift(y, I->shift) + olh->y0;
-    if (x > (int64_t) max_fixed) {
-        x = (int64_t) max_fixed;
-    }
-    else if (x < (int64_t) min_fixed) {
-        x = (int64_t) min_fixed;
+    x = import_shift(x, I->shift);
+    y = -import_shift(y, I->shift);
+    if (olh->fserver->transform_outline) {
+        gs_point pt;
+        I->gs_error = gs_distance_transform((double)fixed2float((float)x), (double)fixed2float((float)y), &olh->fserver->outline_mat, &pt);
+        if (I->gs_error < 0)
+            return I->gs_error;
+        x = float2fixed(pt.x);
+        y = float2fixed(pt.y);
     }
+    x += olh->x0;
+    y += olh->y0;
 
-    if (y > (int64_t) max_fixed) {
-        y = (int64_t) max_fixed;
-    }
-    else if (y < (int64_t) min_fixed) {
-        y = (int64_t) min_fixed;
+    if (x > (int64_t) max_coord_fixed || x < (int64_t) min_coord_fixed
+     || y > (int64_t) max_coord_fixed || y < (int64_t) min_coord_fixed) {
+         I->gs_error = gs_error_undefinedresult;
     }
+    else {
+        olh->need_close = true;
 
-    olh->need_close = true;
-    I->gs_error = gx_path_add_line_notes(olh->path, (fixed) x, (fixed) y, 0);
+/*        dprintf2("%f %f lineto\n", fixed2float(x), fixed2float(y)); */
+        I->gs_error = gx_path_add_line_notes(olh->path, (fixed) x, (fixed) y, 0);
+    }
     return (I->gs_error);
 }
 
@@ -155,57 +171,53 @@ add_curve(gs_fapi_path *I, int64_t x0, int64_t y0, int64_t x1, int64_t y1,
 {
     gs_fapi_outline_handler *olh = (gs_fapi_outline_handler *) I->olh;
 
-    x0 = import_shift(x0, I->shift) + olh->x0;
-    y0 = -import_shift(y0, I->shift) + olh->y0;
-    x1 = import_shift(x1, I->shift) + olh->x0;
-    y1 = -import_shift(y1, I->shift) + olh->y0;
-    x2 = import_shift(x2, I->shift) + olh->x0;
-    y2 = -import_shift(y2, I->shift) + olh->y0;
+    x0 = import_shift(x0, I->shift);
+    y0 = -import_shift(y0, I->shift);
+    x1 = import_shift(x1, I->shift);
+    y1 = -import_shift(y1, I->shift);
+    x2 = import_shift(x2, I->shift);
+    y2 = -import_shift(y2, I->shift);
 
-    if (x0 > (int64_t) max_fixed) {
-        x0 = (int64_t) max_fixed;
-    }
-    else if (x0 < (int64_t) min_fixed) {
-        x0 = (int64_t) min_fixed;
+    if (olh->fserver->transform_outline) {
+        gs_point pt;
+        I->gs_error = gs_distance_transform((double)fixed2float((float)x0), (double)fixed2float((float)y0), &olh->fserver->outline_mat, &pt);
+        if (I->gs_error < 0)
+            return I->gs_error;
+        x0 = float2fixed(pt.x);
+        y0 = float2fixed(pt.y);
+        I->gs_error = gs_distance_transform((double)fixed2float((float)x1), (double)fixed2float((float)y1), &olh->fserver->outline_mat, &pt);
+        if (I->gs_error < 0)
+            return I->gs_error;
+        x1 = float2fixed(pt.x);
+        y1 = float2fixed(pt.y);
+        I->gs_error = gs_distance_transform((double)fixed2float((float)x2), (double)fixed2float((float)y2), &olh->fserver->outline_mat, &pt);
+        if (I->gs_error < 0)
+            return I->gs_error;
+        x2 = float2fixed(pt.x);
+        y2 = float2fixed(pt.y);
     }
+    x0 += olh->x0;
+    y0 += olh->y0;
+    x1 += olh->x0;
+    y1 += olh->y0;
+    x2 += olh->x0;
+    y2 += olh->y0;
 
-    if (y0 > (int64_t) max_fixed) {
-        y0 = (int64_t) max_fixed;
-    }
-    else if (y0 < (int64_t) min_fixed) {
-        y0 = (int64_t) min_fixed;
-    }
-    if (x1 > (int64_t) max_fixed) {
-        x1 = (int64_t) max_fixed;
-    }
-    else if (x1 < (int64_t) min_fixed) {
-        x1 = (int64_t) min_fixed;
+    if (x0 > (int64_t) max_coord_fixed || x0 < (int64_t) min_coord_fixed
+     || y0 > (int64_t) max_coord_fixed || y0 < (int64_t) min_coord_fixed
+     || x1 > (int64_t) max_coord_fixed || x1 < (int64_t) min_coord_fixed
+     || y1 > (int64_t) max_coord_fixed || y1 < (int64_t) min_coord_fixed
+     || x2 > (int64_t) max_coord_fixed || x2 < (int64_t) min_coord_fixed
+     || y2 > (int64_t) max_coord_fixed || y2 < (int64_t) min_coord_fixed)
+    {
+        I->gs_error = gs_error_undefinedresult;
     }
+    else {
+        olh->need_close = true;
 
-    if (y1 > (int64_t) max_fixed) {
-        y1 = (int64_t) max_fixed;
+/*        dprintf6("%f %f %f %f %f %f curveto\n", fixed2float(x0), fixed2float(y0), fixed2float(x1), fixed2float(y1), fixed2float(x2), fixed2float(y2));*/
+        I->gs_error = gx_path_add_curve_notes(olh->path, (fixed) x0, (fixed) y0, (fixed) x1, (fixed) y1, (fixed) x2, (fixed) y2, 0);
     }
-    else if (y1 < (int64_t) min_fixed) {
-        y1 = (int64_t) min_fixed;
-    }
-    if (x2 > (int64_t) max_fixed) {
-        x2 = (int64_t) max_fixed;
-    }
-    else if (x2 < (int64_t) min_fixed) {
-        x2 = (int64_t) min_fixed;
-    }
-
-    if (y2 > (int64_t) max_fixed) {
-        y2 = (int64_t) max_fixed;
-    }
-    else if (y2 < (int64_t) min_fixed) {
-        y2 = (int64_t) min_fixed;
-    }
-
-    olh->need_close = true;
-    I->gs_error =
-        gx_path_add_curve_notes(olh->path, (fixed) x0, (fixed) y0, (fixed) x1,
-                                (fixed) y1, (fixed) x2, (fixed) y2, 0);
     return (I->gs_error);
 }
 
@@ -231,7 +243,7 @@ gs_fapi_get_metrics_count(gs_fapi_font *ff)
  * we know the character is well-behaved, i.e., is not defined by an
  * arbitrary PostScript procedure.
  */
-static bool
+static inline bool
 fapi_gs_char_show_width_only(const gs_text_enum_t *penum)
 {
     if (!gs_text_is_width_only(penum))
@@ -256,14 +268,42 @@ fapi_gs_char_show_width_only(const gs_text_enum_t *penum)
  * transparency is involved - if so, we have to produce
  * a path outline, and not a bitmap.
  */
-static bool
+static inline bool
 using_transparency_pattern(gs_gstate *pgs)
 {
     gx_device *dev = gs_currentdevice_inline(pgs);
 
-    return ((!gs_color_writes_pure(pgs))
-            && dev_proc(dev, begin_transparency_group) != gx_default_begin_transparency_group
-            && dev_proc(dev, end_transparency_group) != gx_default_end_transparency_group);
+    return ((!gs_color_writes_pure(pgs)) && dev_proc(dev, dev_spec_op)(dev, gxdso_supports_pattern_transparency, NULL, 0));
+}
+
+static inline bool
+recreate_multiple_master(gs_font_base *pbfont)
+{
+    bool r = false;
+    gs_fapi_server *I = pbfont->FAPI;
+    bool changed = false;
+
+    if (I && I->face.font_id != gs_no_id &&
+        (pbfont->FontType == ft_encrypted
+        || pbfont->FontType == ft_encrypted2)) {
+        gs_font_type1 *pfont1 = (gs_font_type1 *) pbfont;
+        if (pfont1->data.WeightVector.count != 0
+            && I->face.WeightVector.count != pfont1->data.WeightVector.count) {
+            changed = true;
+        }
+        else if (pfont1->data.WeightVector.count != 0) {
+            changed = (memcmp(I->face.WeightVector.values, pfont1->data.WeightVector.values,
+                             pfont1->data.WeightVector.count * sizeof(pfont1->data.WeightVector.values[0])) != 0);
+        }
+
+        if (changed) {
+            r = (I->set_mm_weight_vector(I, &I->ff, pfont1->data.WeightVector.values, pfont1->data.WeightVector.count) == gs_error_invalidaccess);
+            I->face.WeightVector.count = pfont1->data.WeightVector.count;
+            memcpy(I->face.WeightVector.values, pfont1->data.WeightVector.values,
+                   pfont1->data.WeightVector.count * sizeof(pfont1->data.WeightVector.values[0]));
+        }
+    }
+    return r;
 }
 
 static bool
@@ -398,7 +438,14 @@ gs_fapi_prepare_font(gs_font *pfont, gs_fapi_server *I, int subfont, const char
         < 0)
         return code;
     pbfont->FAPI_font_data = I->ff.server_font_data;    /* Save it back to GS font. */
-    if (I->ff.server_font_data != 0) {
+
+    /* We only want to "refine" the FontBBox for fonts where we allow FAPI to be
+       treated as a "black box", handing over the entire font to the FAPI server.
+       That is, fonts where we give the server either the file, or a buffer with
+       the entire font description in it.
+     */
+    if (I->ff.server_font_data != 0
+        && (font_file_path != NULL || full_font_buf != NULL)) {
         if ((code =
              gs_fapi_renderer_retcode(mem, I,
                                       I->get_font_bbox(I, &I->ff,
@@ -531,26 +578,16 @@ outline_char(gs_memory_t *mem, gs_fapi_server *I, int import_shift_v,
 {
     gs_fapi_path path_interface = path_interface_stub;
     gs_fapi_outline_handler olh;
-    int code;
-    gs_gstate *pgs;
+    int code = 0;
+    gs_gstate *pgs = penum_s->pgs;
+    struct gx_path_s path1;
 
-    extern_st(st_gs_show_enum);
-    extern_st(st_gs_gstate);
+    (void)gx_path_init_local(&path1, mem);
 
-    if (gs_object_type(penum_s->memory, penum_s) == &st_gs_show_enum) {
-        pgs = penum_s->pgs;
-    }
-    else {
-        if (gs_object_type(penum_s->memory, penum_s->pgs) == &st_gs_gstate) {
-            pgs = (gs_gstate *) penum_s->pgs;
-        }
-        else
-            /* No graphics state, give up... */
-            return_error(gs_error_undefined);
-    }
-    olh.path = path;
-    olh.x0 = pgs->ctm.tx_fixed;
-    olh.y0 = pgs->ctm.ty_fixed;
+    olh.fserver = I;
+    olh.path = &path1;
+    olh.x0 = pgs->ctm.tx_fixed - float2fixed(penum_s->fapi_glyph_shift.x);
+    olh.y0 = pgs->ctm.ty_fixed - float2fixed(penum_s->fapi_glyph_shift.y);
     olh.close_path = close_path;
     olh.need_close = false;
     path_interface.olh = &olh;
@@ -560,15 +597,22 @@ outline_char(gs_memory_t *mem, gs_fapi_server *I, int import_shift_v,
                                   I->get_char_outline(I,
                                                       &path_interface))) < 0
         || path_interface.gs_error != 0) {
-        if (path_interface.gs_error != 0)
-            return path_interface.gs_error;
-        else
-            return code;
+        if (path_interface.gs_error != 0) {
+            code = path_interface.gs_error;
+            goto done;
+        }
+        else {
+            goto done;
+        }
     }
     if (olh.need_close && olh.close_path)
         if ((code = add_closepath(&path_interface)) < 0)
-            return code;
-    return 0;
+            goto done;
+    code = gx_path_copy(&path1, path);
+done:
+    code = code >= 0 || code == gs_error_undefinedresult ? 0 : code;
+    gx_path_free(&path1, "outline_char");
+    return code;
 }
 
 static void
@@ -760,7 +804,7 @@ fapi_image_uncached_glyph(gs_font *pfont, gs_gstate *pgs, gs_show_enum *penum,
     gs_gstate *penum_pgs = (gs_gstate *) penum->pgs;
     int code;
     const gx_clip_path *pcpath = pgs->clip_path;
-    const gx_drawing_color *pdcolor = penum->pdcolor;
+    const gx_drawing_color *pdcolor = gs_currentdevicecolor_inline(penum->pgs);
     int rast_orig_x = rast->orig_x;
     int rast_orig_y = -rast->orig_y;
     gs_font_base *pbfont = (gs_font_base *)pfont;
@@ -793,7 +837,7 @@ fapi_image_uncached_glyph(gs_font *pfont, gs_gstate *pgs, gs_show_enum *penum,
              * compression ratios) we therefore set such bytes to zero. It would
              * be nicer if this was fixed in future.
              */
-            r = gs_alloc_bytes(penum->memory, dstr * rast->height,
+            r = gs_alloc_bytes(penum->memory, (size_t)dstr * rast->height,
                                "fapi_finish_render_aux");
             if (!r) {
                 return_error(gs_error_VMerror);
@@ -825,6 +869,15 @@ fapi_image_uncached_glyph(gs_font *pfont, gs_gstate *pgs, gs_show_enum *penum,
             dx += penum->fapi_glyph_shift.x;
             dy += penum->fapi_glyph_shift.y;
         }
+        /* Processing an image object operation, but this may be for a text object */
+        ensure_tag_is_set(pgs, pgs->device, GS_TEXT_TAG);	/* NB: may unset_dev_color */
+        code = gx_set_dev_color(pgs);
+        if (code != 0)
+            return code;
+        code = gs_gstate_color_load(pgs);
+        if (code < 0)
+            return code;
+
         code = gx_image_fill_masked(dev, r, 0, dstr, gx_no_bitmap_id,
                                     (int)dx, (int)dy,
                                     rast->width, rast->height,
@@ -847,7 +900,7 @@ fapi_image_uncached_glyph(gs_font *pfont, gs_gstate *pgs, gs_show_enum *penum,
         byte *bold_lines = NULL;
         byte *line = NULL;
         int ascent = 0;
-        
+
         pie = gs_image_enum_alloc(mem, "image_char(image_enum)");
         if (!pie) {
             return_error(gs_error_VMerror);
@@ -879,7 +932,7 @@ fapi_image_uncached_glyph(gs_font *pfont, gs_gstate *pgs, gs_show_enum *penum,
         image.Width = w + bold;
         image.Height = h + bold;
         image.adjust = false;
-        code = gs_image_init(pie, &image, false, false, penum_pgs);
+        code = gs_image_init(pie, &image, false, true, penum_pgs);
         nbytes = (rast->width + 7) >> 3;
 
         switch (code) {
@@ -918,7 +971,7 @@ fapi_image_uncached_glyph(gs_font *pfont, gs_gstate *pgs, gs_show_enum *penum,
 
                                 bits_merge(merged_line(y - kmask), merged_line(y - (kmask >> 1)), dest_bytes);
                             }
-                        }   
+                        }
 
                        /*
                         * As of this point in the loop, we maintain the following
@@ -976,27 +1029,10 @@ gs_fapi_finish_render(gs_font *pfont, gs_gstate *pgs, gs_text_enum_t *penum, gs_
     gs_memory_t *mem = pfont->memory;
     gs_font_base *pbfont = (gs_font_base *)pfont;
 
-    extern_st(st_gs_show_enum);
-    extern_st(st_gs_gstate);
-
-    if (penum == NULL) {
+    if (penum == NULL)
         return_error(gs_error_undefined);
-    }
 
-    /* Ensure that pgs points to a st_gs_gstate (graphics state) structure */
-    if (gs_object_type(penum->memory, penum->pgs) != &st_gs_gstate) {
-        /* If pgs is not a graphics state, see if the text enumerator is a
-         * show enumerator, in which case we have a pointer to the graphics state there
-         */
-        if (gs_object_type(penum->memory, penum) == &st_gs_show_enum) {
-            penum_pgs = penum_s->pgs;
-        }
-        else
-            /* No graphics state, give up... */
-            return_error(gs_error_undefined);
-    }
-    else
-        penum_pgs = (gs_gstate *) penum->pgs;
+    penum_pgs = penum_s->pgs;
 
     dev1 = gs_currentdevice_inline(penum_pgs);  /* Possibly changed by zchar_set_cache. */
 
@@ -1147,8 +1183,7 @@ gs_fapi_finish_render(gs_font *pfont, gs_gstate *pgs, gs_text_enum_t *penum, gs_
                     if ((code = fapi_copy_mono(dev1, &rast, dx, dy)) < 0)
                         return code;
 
-                    if (gs_object_type(penum->memory, penum) ==
-                        &st_gs_show_enum) {
+                    if (penum_s->cc != NULL) {
                         penum_s->cc->offset.x +=
                             float2fixed(penum_s->fapi_glyph_shift.x);
                         penum_s->cc->offset.y +=
@@ -1221,6 +1256,7 @@ gs_fapi_do_char(gs_font *pfont, gs_gstate *pgs, gs_text_enum_t *penum, char *fon
         return 0;
 
     I->use_outline = false;
+    I->transform_outline = false;
 
     if (penum == 0)
         return_error(gs_error_undefined);
@@ -1243,7 +1279,10 @@ gs_fapi_do_char(gs_font *pfont, gs_gstate *pgs, gs_text_enum_t *penum, char *fon
             (pbfont->dir->ccache.upper >> 1) : MAX_CCACHE_TEMP_BITMAP_BITS;
     }
 
-    I->grid_fit = pbfont->dir->grid_fit_tt;
+    if (pbfont->memory->gs_lib_ctx->font_dir != NULL)
+        I->grid_fit = pbfont->memory->gs_lib_ctx->font_dir->grid_fit_tt;
+    else
+        I->grid_fit = pbfont->dir->grid_fit_tt;
 
     /* Compute the scale : */
     if (!SHOW_IS(penum, TEXT_DO_NONE) && !I->use_outline) {
@@ -1278,8 +1317,18 @@ gs_fapi_do_char(gs_font *pfont, gs_gstate *pgs, gs_text_enum_t *penum, char *fon
     I->ff.is_vertical = bVertical;
     I->ff.client_ctx_p = I->client_ctx_p;
 
-    scale = 1 << I->frac_shift;
-  retry_oversampling:
+    if (recreate_multiple_master(pbfont)) {
+        if ((void *)pbfont->base == (void *)pbfont) {
+           gs_fapi_release_typeface(I, &pbfont->FAPI_font_data);
+        }
+        else {
+            pbfont->FAPI_font_data = NULL;
+            pbfont->FAPI->face.font_id = gs_no_id;
+        }
+    }
+
+   scale = 1 << I->frac_shift;
+retry_oversampling:
     if (I->face.font_id != pbfont->id ||
         !MTX_EQ((&I->face.ctm), ctm) ||
         I->face.log2_scale.x != log2_scale.x ||
@@ -1293,10 +1342,9 @@ gs_fapi_do_char(gs_font *pfont, gs_gstate *pgs, gs_text_enum_t *penum, char *fon
         , {1, 1}
         , true
         };
-        gs_matrix scale_mat, scale_ctm;
 
+        gs_matrix lctm, scale_mat, scale_ctm;
         I->face.font_id = pbfont->id;
-        I->face.ctm = *ctm;
         I->face.log2_scale = log2_scale;
         I->face.align_to_pixels = align_to_pixels;
         I->face.HWResolution[0] = dev->HWResolution[0];
@@ -1309,7 +1357,9 @@ gs_fapi_do_char(gs_font *pfont, gs_gstate *pgs, gs_text_enum_t *penum, char *fon
         /* We apply the entire transform to the glyph (that is ctm x FontMatrix)
          * at render time.
          */
-
+        lctm = *ctm;
+retry_scaling:
+        I->face.ctm = lctm;
         memset(&scale_ctm, 0x00, sizeof(gs_matrix));
         scale_ctm.xx = dev->HWResolution[0] / 72;
         scale_ctm.yy = dev->HWResolution[1] / 72;
@@ -1317,7 +1367,7 @@ gs_fapi_do_char(gs_font *pfont, gs_gstate *pgs, gs_text_enum_t *penum, char *fon
         if ((code = gs_matrix_invert((const gs_matrix *)&scale_ctm, &scale_ctm)) < 0)
             return code;
 
-        if ((code = gs_matrix_multiply(ctm, &scale_ctm, &scale_mat)) < 0)  /* scale_mat ==  CTM - resolution scaling */
+        if ((code = gs_matrix_multiply(&lctm, &scale_ctm, &scale_mat)) < 0)  /* scale_mat ==  CTM - resolution scaling */
             return code;
 
         if ((code = I->get_fontmatrix(I, &scale_ctm)) < 0)
@@ -1329,18 +1379,38 @@ gs_fapi_do_char(gs_font *pfont, gs_gstate *pgs, gs_text_enum_t *penum, char *fon
         if ((code = gs_matrix_multiply(&scale_mat, &scale_ctm, &scale_mat)) < 0)  /* scale_mat ==  CTM - resolution scaling - FontMatrix scaling */
             return code;
 
-        font_scale.matrix[0] =
-            (fracint) (scale_mat.xx * FontMatrix_div * scale + 0.5);
-        font_scale.matrix[1] =
-            -(fracint) (scale_mat.xy * FontMatrix_div * scale + 0.5);
-        font_scale.matrix[2] =
-            (fracint) (scale_mat.yx * FontMatrix_div * scale + 0.5);
-        font_scale.matrix[3] =
-            -(fracint) (scale_mat.yy * FontMatrix_div * scale + 0.5);
-        font_scale.matrix[4] =
-            (fracint) (scale_mat.tx * FontMatrix_div * scale + 0.5);
-        font_scale.matrix[5] =
-            (fracint) (scale_mat.ty * FontMatrix_div * scale + 0.5);
+        if (((int64_t)(scale_mat.xx * FontMatrix_div * scale + 0.5)) != ((int32_t)(scale_mat.xx * FontMatrix_div * scale + 0.5))
+        ||  ((int64_t)(scale_mat.xy * FontMatrix_div * scale + 0.5)) != ((int32_t)(scale_mat.xy * FontMatrix_div * scale + 0.5))
+        ||  ((int64_t)(scale_mat.yx * FontMatrix_div * scale + 0.5)) != ((int32_t)(scale_mat.yx * FontMatrix_div * scale + 0.5))
+        ||  ((int64_t)(scale_mat.yy * FontMatrix_div * scale + 0.5)) != ((int32_t)(scale_mat.yy * FontMatrix_div * scale + 0.5))) {
+            /* Overflow
+               If the scaling is large enough to overflow the 16.16 representation, we forcibly produce an outline
+               unscaled except an arbitrary "midrange" scale (chosen to avoid under/overflow issues). And
+               then scale the points as we create the Ghostscript path outline_char().
+               If the glyph is this large, we're really not worried about hinting or dropout detection etc.
+             */
+
+            memset(&lctm, 0x00, sizeof(gs_matrix));
+            lctm.xx = 256.0;
+            lctm.yy = 256.0;
+            I->transform_outline = true;
+            I->use_outline = true;
+            if ((code = gs_matrix_invert((const gs_matrix *)&lctm, &scale_ctm)) < 0)
+                 return code;
+            if ((code = gs_matrix_multiply(ctm, &scale_ctm, &scale_mat)) < 0)  /* scale_mat ==  CTM - resolution scaling */
+                return code;
+
+            I->outline_mat = scale_mat;
+            goto retry_scaling;
+        }
+        else {
+            font_scale.matrix[0] =  FAPI_ROUND_TO_FRACINT(scale_mat.xx * FontMatrix_div * scale);
+            font_scale.matrix[1] = -FAPI_ROUND_TO_FRACINT(scale_mat.xy * FontMatrix_div * scale);
+            font_scale.matrix[2] =  FAPI_ROUND_TO_FRACINT(scale_mat.yx * FontMatrix_div * scale);
+            font_scale.matrix[3] = -FAPI_ROUND_TO_FRACINT(scale_mat.yy * FontMatrix_div * scale);
+            font_scale.matrix[4] =  FAPI_ROUND_TO_FRACINT(scale_mat.tx * FontMatrix_div * scale);
+            font_scale.matrix[5] =  FAPI_ROUND_TO_FRACINT(scale_mat.ty * FontMatrix_div * scale);
+        }
 
         /* Note: the ctm mapping here is upside down. */
         font_scale.HWResolution[0] =
@@ -1357,6 +1427,7 @@ gs_fapi_do_char(gs_font *pfont, gs_gstate *pgs, gs_text_enum_t *penum, char *fon
                       (double)font_scale.matrix[3]) == 0.0)) {
 
             /* If the matrix is degenerate, force a scale to 1 unit. */
+            memset(&font_scale.matrix, 0x00, sizeof(font_scale.matrix));
             if (!font_scale.matrix[0])
                 font_scale.matrix[0] = 1;
             if (!font_scale.matrix[3])
@@ -1370,6 +1441,7 @@ gs_fapi_do_char(gs_font *pfont, gs_gstate *pgs, gs_text_enum_t *penum, char *fon
                                                          gs_fapi_toplevel_prepared))) < 0) {
             return code;
         }
+        pbfont->FAPI_font_data = I->ff.server_font_data;    /* Save it back to GS font. */
     }
 
     cr.char_codes_count = 1;
@@ -1467,7 +1539,7 @@ gs_fapi_do_char(gs_font *pfont, gs_gstate *pgs, gs_text_enum_t *penum, char *fon
     }
     if (cr.metrics_type == gs_fapi_metrics_notdef && !bVertical) {
         code =
-            I->ff.fapi_get_metrics(&I->ff, &enc_char_name_string, index, sbw,
+            I->ff.fapi_get_metrics(&I->ff, &enc_char_name_string, (int)index, sbw,
                                    bVertical);
         if (code < 0)
             return code;
@@ -1669,15 +1741,15 @@ gs_fapi_do_char(gs_font *pfont, gs_gstate *pgs, gs_text_enum_t *penum, char *fon
 
     if (bCID) {
         code =
-            I->ff.fapi_set_cache(penum, pbfont, &enc_char_name_string, index,
+            I->ff.fapi_set_cache(penum, pbfont, &enc_char_name_string, index + GS_MIN_CID_GLYPH,
                                  sbw + 2, &char_bbox, sbwp, &imagenow);
     }
     else {
         code =
-            I->ff.fapi_set_cache(penum, pbfont, &enc_char_name_string, -1,
+            I->ff.fapi_set_cache(penum, pbfont, &enc_char_name_string, index,
                                  sbw + 2, &char_bbox, sbwp, &imagenow);
     }
-    
+
     /* If we can render the glyph now, do so.
      * we may not be able to in the PS world if there's a CDevProc in the font
      * in which case gs_fapi_finish_render() will be called from the PS
@@ -1728,12 +1800,12 @@ gs_fapi_passfont(gs_font *pfont, int subfont, char *font_file_path,
     int code = 0;
     gs_fapi_server *I, **list;
     bool free_params = false;
-    gs_memory_t *mem = pfont->memory; 
+    gs_memory_t *mem = pfont->memory;
     const char *decodingID = NULL;
     bool do_restart = false;
-    
+
     list = gs_fapi_get_server_list(mem);
-    
+
     if (!list) /* Should never happen */
       return_error(gs_error_unregistered);
 
